from typing import AsyncIterator, cast
from uuid import uuid1

import pytest
from arango.collection import StandardCollection
from arango.database import StandardDatabase

from resotocore.db.async_arangodb import AsyncArangoDB


@pytest.fixture
async def test_collection(test_db: StandardDatabase) -> AsyncIterator[StandardCollection]:
    tmp_name = "tmp_" + str(uuid1()).replace("-", "")
    print(tmp_name)
    try:
        yield cast(StandardCollection, test_db.create_collection(tmp_name))
    finally:
        test_db.delete_collection(tmp_name, ignore_missing=True)


@pytest.mark.asyncio
async def test_events(async_db: AsyncArangoDB, test_collection: StandardCollection) -> None:
    tc = test_collection.name

    with pytest.raises(Exception):
        async with async_db.begin_transaction(read=[tc], write=[tc]) as tx:
            await tx.insert(tc, {"_key": "foo"})
            raise Exception("foo")
    result = list(await async_db.all(tc))
    assert len(result) == 0

    async with async_db.begin_transaction(read=[tc], write=[tc]) as tx:
        await tx.insert(tc, {"_key": "foo"})
    result = list(await async_db.all(tc))
    assert len(result) == 1



import json
from typing import List
from xml.etree import ElementTree

import pytest
import yaml
from aiostream import stream
from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import lists

from resotocore.types import JsonElement, Json
from resotocore.web.content_renderer import (
    respond_json,
    respond_ndjson,
    respond_yaml,
    respond_dot,
    respond_text,
    respond_cytoscape,
    respond_graphml,
)
from tests.resotocore.hypothesis_extension import (
    json_array_gen,
    json_simple_element_gen,
    node_gen,
    graph_stream,
)


@given(json_array_gen)
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_json(elements: List[JsonElement]) -> None:
    async with stream.iterate(elements).stream() as streamer:
        result = ""
        async for elem in respond_json(streamer):
            result += elem
        assert json.loads(result) == elements


@given(json_array_gen)
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_ndjson(elements: List[JsonElement]) -> None:
    async with stream.iterate(elements).stream() as streamer:
        result = []
        async for elem in respond_ndjson(streamer):
            result.append(json.loads(elem.strip()))
        assert result == elements


@given(json_array_gen)
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_yaml(elements: List[JsonElement]) -> None:
    async with stream.iterate(elements).stream() as streamer:
        result = ""
        async for elem in respond_yaml(streamer):
            result += elem + "\n"
        assert [a for a in yaml.full_load_all(result)] == elements


@given(lists(json_simple_element_gen, min_size=1, max_size=10))
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_text_simple_elements(elements: List[JsonElement]) -> None:
    async with stream.iterate(elements).stream() as streamer:
        result = ""
        async for elem in respond_text(streamer):
            result += elem + "\n"
        # every element is rendered as one or more line (string with \n is rendered as multiple lines)
        assert len(elements) + 1 <= len(result.split("\n"))


@given(lists(node_gen(), min_size=1, max_size=10))
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_text_complex_elements(elements: List[JsonElement]) -> None:
    async with stream.iterate(elements).stream() as streamer:
        result = ""
        async for elem in respond_text(streamer):
            result += elem
        # every element is rendered as yaml with --- as object deliminator
        assert len(elements) == len(result.split("---"))


@given(lists(node_gen(), min_size=1, max_size=10))
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_cytoscape(elements: List[Json]) -> None:
    async with graph_stream(elements).stream() as streamer:
        result = ""
        async for elem in respond_cytoscape(streamer):
            result += elem
        # The resulting string can be parsed as json
        assert json.loads(result)


@given(lists(node_gen(), min_size=1, max_size=10))
@settings(max_examples=20, suppress_health_check=list(HealthCheck), deadline=1000)
@pytest.mark.asyncio
async def test_graphml(elements: List[Json]) -> None:
    async with graph_stream(elements).stream() as streamer:
        result = ""
        async for elem in respond_graphml(streamer):
            result += elem
    # The resulting string can be parsed as xml
    assert ElementTree.fromstring(result) is not None


@pytest.mark.asyncio
async def test_dot() -> None:
    def node(name: str, account_name: str) -> Json:
        ancestors = {"account": {"reported": {"name": account_name}}}
        return {"type": "node", "id": name, "reported": {"kind": name, "name": name}, "ancestors": ancestors}

    def edge(from_node: str, to_node: str) -> Json:
        return {"type": "edge", "from": from_node, "to": to_node, "edge_type": "delete"}

    nodes = [node("a", "acc1"), node("b", "acc1"), node("c", "acc2")]
    edges = [edge("a", "b"), edge("a", "c"), edge("b", "c")]

    async with stream.iterate(nodes + edges).stream() as streamer:
        result = ""
        async for elem in respond_dot(streamer):
            result += elem + "\n"
        expected = (
            "digraph {\n"
            "rankdir=LR\n"
            "overlap=false\n"
            "splines=true\n"
            "node [shape=Mrecord colorscheme=paired12]\n"
            "edge [arrowsize=0.5]\n"
            ' "a" [label="a|a", style=filled fillcolor=1];\n'
            ' "b" [label="b|b", style=filled fillcolor=2];\n'
            ' "c" [label="c|c", style=filled fillcolor=3];\n'
            ' "a" -> "b" [label="delete"]\n'
            ' "a" -> "c" [label="delete"]\n'
            ' "b" -> "c" [label="delete"]\n'
            ' subgraph "acc1" {\n'
            '    "a"\n'
            '    "b"\n'
            " }\n"
            ' subgraph "acc2" {\n'
            '    "c"\n'
            " }\n"
            "}\n"
        )
        assert result == expected



from resotocore.db.arangodb_functions import in_subnet, has_desired_change, has_key
from resotocore.db.model import QueryModel
from resotocore.model.model import Model
from resotocore.query.model import FunctionTerm, Query, IsTerm
from resotocore.types import Json


def test_has_desired_change() -> None:
    result = has_desired_change("crs", FunctionTerm("has_desired_change", "foo.bla", []))
    assert result == "crs.desired.foo.bla!=null && crs.reported.foo.bla!=crs.desired.foo.bla"


def test_ip_range() -> None:
    bind_vars: Json = {}
    model = QueryModel(Query.by(IsTerm(["foo"])).on_section("reported"), Model.empty())
    result = in_subnet("crs", bind_vars, FunctionTerm("in_subnet", "foo.bla", ["192.168.1.0/24"]), model)
    assert result == "BIT_AND(IPV4_TO_NUMBER(crs.foo.bla), 4294967040) == @0"
    assert bind_vars["0"] == 3232235776


def test_has_key() -> None:
    bind_vars: Json = {}
    model = QueryModel(Query.by("foo"), Model.empty())
    result = has_key("crs", bind_vars, FunctionTerm("has_key", "foo.bla", [["a", "b", "c"]]), model)
    assert result == "@fn0 ALL IN ATTRIBUTES(crs.foo.bla, true)"
    assert bind_vars["fn0"] == ["a", "b", "c"]
    bind_vars2: Json = {}
    result = has_key("crs", bind_vars2, FunctionTerm("has_key", "foo.bla", ["a"]), model)
    assert result == "HAS(crs.foo.bla, @fn0)"
    assert bind_vars2["fn0"] == "a"



from resotocore.infra_apps.local_runtime import LocalResotocoreAppRuntime
from resotocore.infra_apps.manifest import AppManifest
from resotocore.ids import InfraAppName
import pytest
from resotocore.cli.model import CLI, EmptyContext
from typing import AsyncIterator, Optional
from argparse import Namespace


@pytest.mark.asyncio
async def test_template_generation(cli: CLI) -> None:
    cleanup_untagged = """
            {%- set tags_part = 'not(has_key(tags, ["' + '", "'.join(config["tags"]) + '"]))' %}
            {%- set kinds_part = 'is(["' + '", "'.join(config["kinds"]) + '"])' %}
            {%- set account_parts = [] %}
            {%- set default_age = config["default"]["age"]|default("2h") %}
            {%- for cloud_id, account in config["accounts"].items() %}
                {%- for account_id, account_data in account.items() %}
                    {%- set age = account_data.get("age", default_age) %}
                    {%- set account_part = '(/ancestors.cloud.reported.id == "' ~ cloud_id ~ '" and /ancestors.account.reported.id == "' ~ account_id ~ '" and age > ' ~ age ~ ')' %}
                    {%- do account_parts.append(account_part) %}
                {%- endfor %}
            {%- endfor %}
            {%- set accounts_part = "(" + " or ".join(account_parts) + ")" %}
            {%- set exclusion_part = "/metadata.protected == false and /metadata.phantom == false and /metadata.cleaned == false" %}
            {%- set required_tags = ", ".join(config["tags"]) %}
            {%- set reason = "Missing one or more of required tags " ~ required_tags ~ " and age more than threshold" %}
            {%- set cleanup_search = 'search ' ~ exclusion_part ~ ' and ' ~ kinds_part ~ ' and ' ~ tags_part ~ ' and ' ~ accounts_part ~ ' | clean "' ~ reason ~ '"' %}
            {{ cleanup_search }}
        """

    config = {
        "tags": ["resoto:cleanup"],
        "kinds": ["instance"],
        "default": {"age": "1s"},
        "accounts": {
            "digitalocean": {
                "12345": {"age": "1s"},
            },
        },
    }

    manifest = AppManifest(
        name=InfraAppName("test-app"),
        description="test app description",
        version="0.0.0",
        readme="",
        language="jinja2",
        url="",
        icon="",
        categories=[],
        default_config=None,
        config_schema=None,
        args_schema={},
        source=cleanup_untagged,
    )

    runtime = LocalResotocoreAppRuntime(cli)
    g = cli.env["graph"]

    lines = [line async for line in runtime.generate_template(g, manifest, config, stdin(), [])]
    assert lines == [
        (
            "search /metadata.protected == false and /metadata.phantom == false and /metadata.cleaned == false "
            'and is(["instance"]) and not(has_key(tags, ["resoto:cleanup"])) and ((/ancestors.cloud.reported.id'
            ' == "digitalocean" and /ancestors.account.reported.id == "12345" and age > 1s)) | '
            'clean "Missing one or more of required tags resoto:cleanup and age more than threshold"'
        ),
    ]


@pytest.mark.asyncio
async def test_execute(cli: CLI) -> None:
    source = 'echo foo; echo bar; json ["a", "b", "c"]'
    manifest = AppManifest(
        name=InfraAppName("test-app"),
        description="test app description",
        version="0.0.0",
        readme="",
        language="jinja2",
        url="",
        icon="",
        categories=[],
        default_config=None,
        config_schema=None,
        args_schema={},
        source=source,
    )

    runtime = LocalResotocoreAppRuntime(cli)
    g = cli.env["graph"]
    output = [
        js_elem async for js_elem in runtime.execute(g, manifest, config={}, argv=[], stdin=stdin(), ctx=EmptyContext)
    ]
    assert output == ["foo", "bar", "a", "b", "c"]


@pytest.mark.asyncio
async def test_search(cli: CLI) -> None:
    source = """
    {% for resource in search('is(foo)') %}
    {{ resource.id }}
    {% endfor %}
    """

    manifest = AppManifest(
        name=InfraAppName("test-app"),
        description="test app description",
        version="0.0.0",
        readme="",
        language="jinja2",
        url="",
        icon="",
        categories=[],
        default_config=None,
        config_schema=None,
        args_schema={},
        source=source,
    )

    runtime = LocalResotocoreAppRuntime(cli)

    g = cli.env["graph"]
    lines = [line async for line in runtime.generate_template(g, manifest, {}, stdin(), [])]

    assert lines == ["sub_root", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]


async def stdin() -> AsyncIterator[Optional[str]]:
    yield None


namespace = Namespace()



import pytest
from hypothesis import given, settings, HealthCheck

from resotocore.model.graph_access import Direction
from resotocore.query.model import (
    P,
    Query,
    AllTerm,
    IsTerm,
    PathRoot,
    Part,
    MergeQuery,
    Navigation,
    MergeTerm,
    NavigateUntilRoot,
    IdTerm,
    FulltextTerm,
    Predicate,
    NotTerm,
    FunctionTerm,
    Limit,
)
from resotocore.query.query_parser import parse_query
from tests.resotocore.query import query


def simple_reference() -> None:
    # only kind
    Query.by("ec2")

    # equality
    Query.by(P.of_kind("ec2") & (P("simple") == "hallo"))
    Query.by(P.of_kind("ec2") & (P("simple") != "hallo"))

    # regex
    Query.by(P.of_kind("ec2") & P("simple").matches("^some.regex[a-d]+$"))
    Query.by(P.of_kind("ec2") & P("simple").not_matches("^some.regex[a-d]+$"))

    # comparator
    Query.by(P.of_kind("ec2") & (P("num") > 23))
    Query.by(P.of_kind("ec2") & (P("num") >= 23))
    Query.by(P.of_kind("ec2") & (P("num") == 23))
    Query.by(P.of_kind("ec2") & (P("num") <= 23))
    Query.by(P.of_kind("ec2") & (P("num") < 23))

    # in set
    Query.by(P.of_kind("ec2") & P("num").is_in([1, 2, 5]))
    Query.by(P.of_kind("ec2") & P("num").is_not_in([1, 2, 5]))

    # array: all above operators are available
    Query.by(P.of_kind("ec2") & (P.array("some.array").for_all() > 12.23))
    Query.by(P.of_kind("ec2") & (P.array("some.array").for_any().is_in([1, 2, 3])))
    Query.by(P.of_kind("ec2") & (P.array("some.array").for_none() == 5))

    # call a function
    Query.by(P.function("in_subnet").on("ip", "1.2.3.4/16"))

    # refine with multiple predicates (all predicates have to match)
    Query.by(P.of_kind("ec2") & P("a").ge(1), P("b") == 2, P("c").matches("aaa"))


def test_simple_query() -> None:
    a = (
        Query.by("ec2", P("cpu") > 4, (P("mem") < 23) | (P("mem") < 59))
        .traverse_out()
        .filter(P("some.int.value") < 1, P("some.other") == 23)
        .traverse_out()
        .filter(P("active") == 12, P.function("in_subnet").on("ip", "1.2.3.4/32"))
    )

    assert (
        str(a) == '((is("ec2") and cpu > 4) and (mem < 23 or mem < 59)) -default-> '
        "(some.int.value < 1 and some.other == 23) -default-> "
        '(active == 12 and in_subnet(ip, "1.2.3.4/32"))'
    )


def test_simplify() -> None:
    # some_criteria | all => all
    assert str((IsTerm(["test"]) | AllTerm())) == "all"
    # some_criteria & all => some_criteria
    assert str((IsTerm(["test"]) & AllTerm())) == 'is("test")'
    # also works in nested setup
    q = Query.by(AllTerm() & ((P("test") == True) & (IsTerm(["test"]) | AllTerm())))
    assert (str(q)) == "test == true"


def test_combine() -> None:
    query1 = Query.by(P("test") == True).traverse_out().combine(Query.by("foo")).combine(Query.by("bla"))
    assert str(query1) == 'test == true -default-> (is("foo") and is("bla"))'
    query2 = (
        Query.by(P("test") == True)
        .traverse_out()
        .combine(Query.by("foo").traverse_out())
        .combine(Query.by("bla").traverse_out())
    )
    assert str(query2) == 'test == true -default-> is("foo") -default-> is("bla") -default->'
    query3 = (
        Query.by(P("test") == True)
        .traverse_out()
        .filter("boo")
        .traverse_out()
        .filter("bar")
        .combine(Query.by("foo"))
        .combine(Query.by("bla"))
    )
    assert str(query3) == 'test == true -default-> is("boo") -default-> ((is("bar") and is("foo")) and is("bla"))'
    query4 = Query.by("a").with_limit(10).combine(Query.by("b").with_limit(2))
    assert query4.current_part.limit == Limit(0, 2)  # minimum is taken
    with pytest.raises(AttributeError):
        # can not combine 2 aggregations
        parse_query("aggregate(sum(1)): is(a)").combine(parse_query("aggregate(sum(1)): is(a)"))
    with pytest.raises(AttributeError):
        # can not combine 2 with statements
        parse_query("is(foo) with(empty, -default->)").combine(parse_query("is(bla) with(empty, -default->)"))
    # combining merge term queries needs special handling
    sq1 = parse_query("age > 23h").on_section()
    mq1 = parse_query("is(foo) {bla: --> is(bla)} bla.test=2").on_section()
    mq2 = parse_query("name==test {foo: --> is(foo)} foo.test=3").on_section()
    assert str(mq1.combine(mq2)) == (
        '(is("foo") and name == "test") '
        '{bla: all -default-> is("bla"), foo: all -default-> is("foo")} '
        "(bla.test == 2 and foo.test == 3)"
    )
    assert str(mq2.combine(mq1)) == (
        '(name == "test" and is("foo")) '
        '{foo: all -default-> is("foo"), bla: all -default-> is("bla")} '
        "(foo.test == 3 and bla.test == 2)"
    )
    assert str(mq1.combine(sq1)) == '(is("foo") and age > "23h") {bla: all -default-> is("bla")} bla.test == 2'
    assert str(sq1.combine(mq1)) == '(is("foo") and age > "23h") {bla: all -default-> is("bla")} bla.test == 2'
    assert str(mq2.combine(sq1)) == '(name == "test" and age > "23h") {foo: all -default-> is("foo")} foo.test == 3'
    assert str(sq1.combine(mq2)) == '(name == "test" and age > "23h") {foo: all -default-> is("foo")} foo.test == 3'


def test_on_section() -> None:
    query = parse_query(
        "aggregate(foo, bla, bar: sum(a) as a, sum(b) as b, sum(1) as c):"
        '(cpu > 4 and (mem < 23 or mem < 59)) with(any, <-- name == "test") sort mem asc --> '
        "(a < 1 and /metadata.b == 23) sort foo asc"
    )
    on_section = (
        "aggregate(r.foo, r.bla, r.bar: sum(r.a) as a, sum(r.b) as b, sum(1) as c):"
        '(r.cpu > 4 and (r.mem < 23 or r.mem < 59)) with(any, <-default- r.name == "test") sort r.mem asc -default-> '
        "(r.a < 1 and metadata.b == 23) sort r.foo asc"
    )
    with_section_r = query.on_section("r")
    # all variables are now prefixed with the section name
    assert str(with_section_r) == on_section
    # all variables that are prefixed with the section name have the section name removed -> reverse operation
    assert str(with_section_r.relative_to_section("r")) == str(query)
    # a query on section root does not change the query
    assert str(with_section_r.on_section(PathRoot)) == on_section
    # a query relative to section root does not change the query
    assert str(with_section_r.relative_to_section(PathRoot)) == on_section


def test_rewrite_usage() -> None:
    def check(query_str: str, expect_before: str, expect_after: str) -> None:
        before, after = parse_query(query_str).current_part.term.split_by_usage()
        assert str(before) == expect_before
        assert str(after) == expect_after

    # filter on a,b,c before the usage is evaluated
    check("usage.cpu.max>3 and ((a<1 and b>1) or c==3)", "((a < 1 and b > 1) or c == 3)", "usage.cpu.max > 3")
    # same query, but different order
    check("((a<1 and b>1) or c==3) and usage.cpu.max>3", "((a < 1 and b > 1) or c == 3)", "usage.cpu.max > 3")
    # usage on top level in combination with or can not be optimized
    check("usage.cpu.max>3 or ((a<1 and b>1) or c==3)", "all", "(usage.cpu.max > 3 or ((a < 1 and b > 1) or c == 3))")


def test_rewrite_ancestors_descendants() -> None:
    # a query without ancestor/descendants is not changed
    assert str(parse_query("(a<1 and b>1) or c==3")) == "((a < 1 and b > 1) or c == 3)"
    # a query with resolved ancestor is not changed
    assert (
        str(parse_query('a<1 and ancestors.cloud.reported.name=="test"').on_section())
        == '(a < 1 and ancestors.cloud.reported.name == "test")'
    )
    # the merge name is interpreted relative to the section
    assert (
        str(parse_query("a<1 {test: <-[1:]- is(account)}").on_section("reported"))
        == 'reported.a < 1 {reported.test: all <-default[1:]- is("account")}'
    )
    # the merge name is not interpreted relative to the section, when defined absolute
    assert (
        str(parse_query("a<1 {/test: <-[1:]- is(account)}").on_section("reported"))
        == 'reported.a < 1 {test: all <-default[1:]- is("account")}'
    )
    # a query with unknown ancestor creates a merge query
    assert (
        str(parse_query('a<1 and ancestors.cloud.reported.kind=="cloud"').on_section())
        == 'a < 1 {ancestors.cloud: all <-default[1:]- is("cloud")} ancestors.cloud.reported.kind == "cloud"'
    )
    # multiple ancestors are put into one merge query
    assert (
        str(
            parse_query(
                'a<1 and ancestors.cloud.reported.kind=="c" and ancestors.account.reported.kind=="a"'
            ).on_section()
        )
        == 'a < 1 {ancestors.cloud: all <-default[1:]- is("cloud"), '
        'ancestors.account: all <-default[1:]- is("account")} '
        '(ancestors.cloud.reported.kind == "c" and ancestors.account.reported.kind == "a")'
    )
    # existing merge queries are preserved
    assert (
        str(parse_query('a<1 {children[]: --> all} ancestors.cloud.reported.kind=="c"').on_section())
        == 'a < 1 {ancestors.cloud: all <-default[1:]- is("cloud"), children[]: all -default-> all} '
        'ancestors.cloud.reported.kind == "c"'
    )
    # predefined merge queries are preserved
    assert (
        str(parse_query('a<1 {ancestors.cloud: --> is(region)} ancestors.cloud.reported.kind=="c"').on_section())
        == 'a < 1 {ancestors.cloud: all -default-> is("region")} ancestors.cloud.reported.kind == "c"'
    )
    # This is an example of a horrible query: all entries have to be merged, before a filter can be applied
    assert (
        str(parse_query("(a<1 and b>1) or ancestors.d.c<1").on_section())
        == 'all {ancestors.d: all <-default[1:]- is("d")} ((a < 1 and b > 1) or ancestors.d.c < 1)'
    )
    # Test some special examples
    assert (
        str(parse_query("ancestors.d.c<1 and (a<1 or b>1) and /ancestors.a.b>1").on_section())
        == '(a < 1 or b > 1) {ancestors.d: all <-default[1:]- is("d"), ancestors.a: all <-default[1:]- is("a")} '
        "(ancestors.d.c < 1 and ancestors.a.b > 1)"
    )
    # the independent query terms are always in the pre-filter before the merge is applied
    assert (
        str(parse_query("(a<1 and b>1) and (c<d or /ancestors.d.c<1)").on_section())
        == str(parse_query("(c<d or /ancestors.d.c<1) and (a<1 and b>1)").on_section())
        == '(a < 1 and b > 1) {ancestors.d: all <-default[1:]- is("d")} (c < "d" or ancestors.d.c < 1)'
    )
    # multiple filters to the same kind only create one merge query
    assert (
        str(
            parse_query(
                "/ancestors.a.b<1 and ancestors.a.c>1 and ancestors.a.d=3 and ancestors.b.c>1 and a==1"
            ).on_section()
        )
        == 'a == 1 {ancestors.a: all <-default[1:]- is("a"), ancestors.b: all <-default[1:]- is("b")} '
        "(((ancestors.a.b < 1 and ancestors.a.c > 1) and ancestors.a.d == 3) and ancestors.b.c > 1)"
    )
    # aggregation queries with ancestors in the group variable trigger a merge
    assert (
        str(parse_query("aggregate(/ancestors.a.reported.name as a: sum(1)): is(volume)").on_section())
        == 'aggregate(ancestors.a.reported.name as a: sum(1)):is("volume") {ancestors.a: all <-default[1:]- is("a")}'
    )


def test_aggregation() -> None:
    q = parse_query('aggregate("{a.a}_{a.b}" as a, a.c.d as v: sum(a.c.e) as c): all')
    assert q.aggregate.property_paths() == {"a.a", "a.b", "a.c.d", "a.c.e"}  # type: ignore


def test_merge_query_creation() -> None:
    inbound = Navigation(1, Navigation.Max, direction=Direction.inbound)
    for_foo = Query([Part(IsTerm(["foo"])), Part(AllTerm(), navigation=inbound)])
    merge_foo = [MergeQuery("ancestors.foo", for_foo)]

    # merge_foo is created automatically
    assert Part(AllTerm()).merge_queries_for(["ancestors.foo.reported.bla"]) == merge_foo
    # merge_foo is already included and not added
    assert Part(MergeTerm(AllTerm(), merge_foo)).merge_queries_for(["ancestors.foo.reported.bla"]) == merge_foo
    # neither ancestors/descendants
    with pytest.raises(Exception):
        Part(AllTerm()).merge_queries_for(["unknown.foo.reported.bla"])
    # no path is given
    with pytest.raises(Exception):
        Part(AllTerm()).merge_queries_for(["ancestors.foo"])

    # rewrite for ancestors/descendants also work with additional properties
    assert (
        str(Query.by("test").rewrite_for_ancestors_descendants(["ancestors.kind.reported.prop", "test", "a"]))
        == 'is("test") {ancestors.kind: all <-default[1:]- is("kind")}'
    )
    assert (
        str(
            Query.by("test")
            .merge_with("ancestors.cloud", NavigateUntilRoot, IsTerm(["cloud"]))
            .rewrite_for_ancestors_descendants(["ancestors.kind.reported.prop", "test", "a"])
        )
        == 'is("test") {ancestors.kind: all <-default[1:]- is("kind"), ancestors.cloud: all <-default[1:]- is("cloud")}'
    )


def test_term_contains() -> None:
    term = parse_query('("test" or "fest") and (p>1 or p<2) {a: <-- is(foo)} not(a>23)').parts[0].term
    assert term.contains_term_type(IdTerm) is False
    assert term.contains_term_type(IsTerm) is True
    assert term.contains_term_type(FulltextTerm) is True
    assert term.contains_term_type(Predicate) is True
    assert term.contains_term_type(NotTerm) is True
    assert term.contains_term_type(FunctionTerm) is False


def test_context_predicates() -> None:
    query: Query = parse_query("a.b[*].{ a=2 and b[1].bla=3 and c.d[*].{ e=4 and f=5 } }")
    expected = ["a.b[*].a", "a.b[*].b[1].bla", "a.b[*].c.d[*].e", "a.b[*].c.d[*].f"]
    assert [str(a.name) for a in query.visible_predicates] == expected


def test_merge_term_combination() -> None:
    sq1 = parse_query("age > 23h").on_section().parts[0].term
    mq1 = parse_query("is(foo) {bla: --> is(bla)} bla.test=2").on_section().parts[0].term
    mq2 = parse_query("name==test {foo: --> is(foo)} foo.test=3").on_section().parts[0].term
    assert (
        str(mq1 | mq2) == '(is("foo") or name == "test") '
        '{bla: all -default-> is("bla"), foo: all -default-> is("foo")} '
        "(bla.test == 2 or foo.test == 3)"
    )
    assert (
        str(mq2 | mq1) == '(name == "test" or is("foo")) '
        '{foo: all -default-> is("foo"), bla: all -default-> is("bla")} '
        "(foo.test == 3 or bla.test == 2)"
    )
    assert (
        str(mq1 & mq2) == '(is("foo") and name == "test") '
        '{bla: all -default-> is("bla"), foo: all -default-> is("foo")} '
        "(bla.test == 2 and foo.test == 3)"
    )
    assert (
        str(mq2 & mq1) == '(name == "test" and is("foo")) '
        '{foo: all -default-> is("foo"), bla: all -default-> is("bla")} '
        "(foo.test == 3 and bla.test == 2)"
    )
    assert str(mq1 | sq1) == '(is("foo") or age > "23h") {bla: all -default-> is("bla")} bla.test == 2'
    assert str(sq1 | mq1) == '(is("foo") or age > "23h") {bla: all -default-> is("bla")} bla.test == 2'
    assert str(mq1 & sq1) == '(is("foo") and age > "23h") {bla: all -default-> is("bla")} bla.test == 2'
    assert str(sq1 & mq1) == '(is("foo") and age > "23h") {bla: all -default-> is("bla")} bla.test == 2'


@given(query)
@settings(max_examples=200, suppress_health_check=list(HealthCheck))
def test_generated_query(q: Query) -> None:
    assert q.structure()



import asyncio
import logging
from asyncio import Task
from contextlib import suppress
from functools import partial
from typing import Optional, List, Callable, Awaitable
from apscheduler.triggers.cron import CronTrigger

import yaml

from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.config import ConfigHandler, ConfigEntity, ConfigValidation
from resotocore.core_config import (
    CoreConfig,
    ResotoCoreConfigId,
    EditableConfig,
    ResotoCoreRoot,
    ResotoCoreCommandsConfigId,
    ResotoCoreCommandsRoot,
    ResotoCoreSnapshotsConfigId,
    ResotoCoreSnapshotsRoot,
    SnapshotsScheduleConfig,
    CustomCommandsConfig,
    migrate_core_config,
    config_model as core_config_model,
    migrate_command_config,
)
from resotocore.system_start import empty_config
from resotocore.ids import SubscriberId, WorkerId, ConfigId
from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.model.model import Kind
from resotocore.model.typed_model import from_js
from resotocore.report import ResotoReportBenchmark, ResotoReportCheck, Inspector, BenchmarkConfigRoot, CheckConfigRoot
from resotocore.report.report_config import config_model as report_config_model
from resotocore.service import Service
from resotocore.types import Json
from resotocore.user import config_model as user_config_model, UsersConfigId, ResotoUsersConfig
from resotocore.util import deep_merge, restart_service, value_in_path, value_in_path_get
from resotocore.worker_task_queue import WorkerTaskQueue, WorkerTaskDescription, WorkerTaskName, WorkerTask

log = logging.getLogger(__name__)


class CoreConfigHandler(Service):
    def __init__(
        self,
        config: CoreConfig,
        message_bus: MessageBus,
        worker_task_queue: WorkerTaskQueue,
        config_handler: ConfigHandler,
        event_sender: AnalyticsEventSender,
        inspector: Inspector,
        exit_fn: Callable[[], None] = partial(restart_service, "resotocore config changed."),
    ):
        self.message_bus = message_bus
        self.worker_task_queue = worker_task_queue
        self.config_updated_listener: Optional[Task[None]] = None
        self.config_validator: Optional[Task[None]] = None
        self.config = config
        self.config_handler = config_handler
        self.event_sender = event_sender
        self.inspector = inspector
        self.exit_fn = exit_fn
        self.config_updated_callbacks: List[Callable[[ConfigId], Awaitable[None]]] = []

    async def validate_config_entry(self, task_data: Json) -> Optional[Json]:
        def validate_core_config() -> Optional[Json]:
            config = value_in_path(task_data, ["config", ResotoCoreRoot])
            if isinstance(config, dict):
                # try to read editable config, throws if there are errors
                read = from_js(config, EditableConfig)
                return read.validate()
            else:
                return {"error": "Expected a json object"}

        def validate_commands_config() -> Optional[Json]:
            config = value_in_path(task_data, ["config", ResotoCoreCommandsRoot])
            if isinstance(config, dict):
                # try to read editable config, throws if there are errors
                read = from_js(config, CustomCommandsConfig)
                return read.validate()
            else:
                return {"error": "Expected a json object"}

        def validate_snapshot_schedule() -> Optional[Json]:
            config = value_in_path(task_data, ["config", ResotoCoreSnapshotsRoot])
            if isinstance(config, dict):
                config = {"snapshots": config}  # wrap in snapshots objects to match the dataclass
                read = from_js(config, SnapshotsScheduleConfig)
                # validate cron expressions
                try:
                    for schedule in read.snapshots.values():
                        CronTrigger.from_crontab(schedule.schedule)
                    return None
                except Exception as ex:
                    return {"error": f"Invalid cron expression: {ex}"}
            else:
                return {"error": "Expected a json object"}

        holder = value_in_path(task_data, ["config"])
        if not isinstance(holder, dict):
            return {"error": "Expected a json object in config"}
        elif ResotoCoreRoot in holder:
            return validate_core_config()
        elif ResotoCoreCommandsRoot in holder:
            return validate_commands_config()
        elif CheckConfigRoot in holder:
            return await self.inspector.validate_check_collection_config(task_data["config"])
        elif BenchmarkConfigRoot in holder:
            return await self.inspector.validate_benchmark_config(task_data["config"])
        elif ResotoCoreSnapshotsRoot in holder:
            return validate_snapshot_schedule()
        else:
            return {"error": "No known configuration found"}

    def add_callback(self, callback: Callable[[ConfigId], Awaitable[None]]) -> None:
        self.config_updated_callbacks.append(callback)

    async def __validate_config(self) -> None:
        worker_id = WorkerId("resotocore.config.validate")
        description = WorkerTaskDescription(
            WorkerTaskName.validate_config,
            {
                "config_id": [
                    ResotoCoreConfigId,
                    ResotoCoreCommandsConfigId,
                    ResotoReportBenchmark,
                    ResotoReportCheck,
                    ResotoCoreSnapshotsConfigId,
                ]
            },
        )
        async with self.worker_task_queue.attach(worker_id, [description]) as tasks:
            while True:
                task: WorkerTask = await tasks.get()
                try:
                    errors = await self.validate_config_entry(task.data)
                    if errors:
                        message = "Validation Errors:\n" + yaml.safe_dump(errors)
                        await self.worker_task_queue.error_task(worker_id, task.id, message)
                    else:
                        await self.worker_task_queue.acknowledge_task(worker_id, task.id)
                except Exception as ex:
                    log.warning("Error processing validate configuration task", exc_info=ex)
                    await self.worker_task_queue.error_task(worker_id, task.id, str(ex))

    async def __detect_usage_metrics_turned_off(self) -> None:
        loaded = await self.config_handler.get_config(ResotoCoreConfigId)
        if loaded:
            enabled = value_in_path_get(loaded.config, [ResotoCoreRoot, "runtime", "usage_metrics"], True)
            if self.config.runtime.usage_metrics and not enabled:
                await self.event_sender.core_event(CoreEvent.UsageMetricsTurnedOff)

    async def __handle_events(self) -> None:
        subscriber_id = SubscriberId("resotocore.config.update")
        async with self.message_bus.subscribe(subscriber_id, [CoreMessage.ConfigUpdated]) as events:
            while True:
                event = await events.get()
                event_id = event.data.get("id")
                if event_id:
                    for callback in self.config_updated_callbacks:
                        try:
                            await callback(event_id)
                        except Exception as ex:
                            log.warning("Error in config update callback", exc_info=ex)

                if event_id in (ResotoCoreConfigId, ResotoCoreCommandsConfigId):
                    log.info(f"Core config was updated: {event_id} Restart to take effect.")
                    await self.__detect_usage_metrics_turned_off()
                    # stop the process and rely on os to restart the service
                    self.exit_fn()

    async def __update_config(self) -> None:
        # in case the internal configuration holds new properties, we update the existing config always.
        try:
            existing = await self.config_handler.get_config(ResotoCoreConfigId)
            empty = empty_config().json()
            updated = deep_merge(empty, existing.config) if existing else empty
            updated = migrate_core_config(updated)
            if existing is None or updated != existing.config:
                await self.config_handler.put_config(ConfigEntity(ResotoCoreConfigId, updated), validate=False)
                log.info("Default resoto config updated.")
        except Exception as ex:
            log.error(f"Could not update resoto default configuration: {ex}", exc_info=ex)

        # make sure there is a default command configuration
        # note: this configuration is only created one time and never updated
        try:
            existing_commands = await self.config_handler.get_config(ResotoCoreCommandsConfigId)
            to_update: Optional[Json] = None
            if existing_commands is None:
                to_update = CustomCommandsConfig().json()
            else:
                to_update = migrate_command_config(existing_commands.config)
            if to_update is not None:
                await self.config_handler.put_config(
                    ConfigEntity(ResotoCoreCommandsConfigId, to_update), validate=False
                )
                log.info("Default resoto commands config updated.")
        except Exception as ex:
            log.error(f"Could not update resoto command configuration: {ex}", exc_info=ex)

        # make sure there is a default user configuration
        try:
            existing_users = await self.config_handler.get_config(UsersConfigId)
            user_update: Optional[Json] = None
            if existing_users is None:
                user_update = ResotoUsersConfig().json()
            if user_update is not None:
                await self.config_handler.put_config(ConfigEntity(UsersConfigId, user_update), validate=False)
                log.info("Default resoto users config updated.")
        except Exception as ex:
            log.error(f"Could not update resoto users configuration: {ex}", exc_info=ex)

        # make sure there is a default snapshots configuration
        try:
            existing_snapshots = await self.config_handler.get_config(ResotoCoreSnapshotsConfigId)
            snapshot_update: Optional[Json] = None
            if existing_snapshots is None:
                snapshot_update = SnapshotsScheduleConfig().json()
            if snapshot_update is not None:
                await self.config_handler.put_config(
                    ConfigEntity(ResotoCoreSnapshotsConfigId, snapshot_update), validate=False
                )
                log.info("Default resoto snapshots config updated.")
        except Exception as ex:
            log.error(f"Could not update resoto snapshots configuration: {ex}", exc_info=ex)

    async def __update_model(self) -> None:
        try:
            models = core_config_model() + report_config_model() + user_config_model()
            kinds = from_js(models, List[Kind])
            await self.config_handler.update_configs_model(kinds)
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoCoreConfigId, external_validation=True)
            )
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoCoreCommandsConfigId, external_validation=True)
            )
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoCoreCommandsConfigId, external_validation=True)
            )
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoCoreSnapshotsConfigId, external_validation=True)
            )
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoReportBenchmark, external_validation=True)
            )
            await self.config_handler.put_config_validation(
                ConfigValidation(ResotoReportCheck, external_validation=True)
            )

            log.debug("Resoto core config model updated.")
        except Exception as ex:
            log.error(f"Could not update resoto core config model: {ex}", exc_info=ex)

    async def start(self) -> None:
        await self.__update_model()
        await self.__update_config()
        self.config_updated_listener = asyncio.create_task(self.__handle_events())
        self.config_validator = asyncio.create_task(self.__validate_config())

    async def stop(self) -> None:
        # wait for the spawned task to complete
        if self.config_updated_listener:
            with suppress(Exception):
                self.config_updated_listener.cancel()
        if self.config_validator:
            with suppress(Exception):
                self.config_validator.cancel()



import collections
import re
from datetime import date
from typing import Optional, cast, Dict

import jsons
import pytest
from deepdiff import DeepDiff
from networkx import MultiDiGraph
from pytest import fixture

from resotocore.ids import NodeId
from resotocore.model.graph_access import GraphAccess, GraphBuilder, EdgeTypes, EdgeKey
from resotocore.model.model import Model, AnyKind, ComplexKind
from resotocore.model.typed_model import to_json
from resotocore.types import Json, EdgeType
from resotocore.util import AccessJson, AccessNone, uuid_str
from tests.resotocore.db.graphdb_test import Foo

FooTuple = collections.namedtuple(
    "FooTuple",
    ["a", "b", "c", "d", "e", "f", "g", "kind"],
    defaults=["", 0, [], "foo", {"a": 12, "b": 32}, date.fromisoformat("2021-03-29"), 1.234567, "foo"],
)


# noinspection PyArgumentList
@fixture
def graph_access() -> GraphAccess:
    g = MultiDiGraph()

    def add_edge(from_node: str, to_node: str, edge_type: EdgeType) -> None:
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        g.add_edge(from_node, to_node, key, edge_type=edge_type)

    g.add_node("1", reported=to_json(FooTuple("1")), desired={"name": "a"}, metadata={"version": 1}, kinds=["foo"])
    g.add_node("2", reported=to_json(FooTuple("2")), desired={"name": "b"}, metadata={"version": 2}, kinds=["foo"])
    g.add_node("3", reported=to_json(FooTuple("3")), desired={"name": "c"}, metadata={"version": 3}, kinds=["foo"])
    g.add_node("4", reported=to_json(FooTuple("4")), desired={"name": "d"}, metadata={"version": 4}, kinds=["foo"])
    add_edge("1", "2", edge_type=EdgeTypes.default)
    add_edge("1", "3", edge_type=EdgeTypes.default)
    add_edge("2", "3", edge_type=EdgeTypes.default)
    add_edge("2", "4", edge_type=EdgeTypes.default)
    add_edge("3", "4", edge_type=EdgeTypes.default)
    add_edge("1", "2", edge_type=EdgeTypes.delete)
    add_edge("1", "3", edge_type=EdgeTypes.delete)
    add_edge("1", "4", edge_type=EdgeTypes.delete)
    return GraphAccess(g)


# noinspection PyArgumentList
def test_access_node() -> None:
    g = MultiDiGraph()
    g.add_node("1", reported=to_json(FooTuple(a="1")))
    access: GraphAccess = GraphAccess(g)
    elem: Json = node(access, "1")  # type: ignore
    assert elem["hash"] == "153c1a5c002f6213a95383f33b63aa18b8ed6939f57418fb0f27312576f0cea4"
    assert elem["reported"] == {
        "a": "1",
        "b": 0,
        "c": [],
        "d": "foo",
        "e": {"a": 12, "b": 32},
        "f": "2021-03-29",
        "g": 1.234567,
        "kind": "foo",
    }
    assert access.node(NodeId("2")) is None


def test_marshal_unmarshal() -> None:
    foo = Foo("12")
    name = type(foo).__name__
    clazz = globals()[name]
    js = jsons.dumps(foo)
    again = jsons.loads(js, cls=clazz)
    assert DeepDiff(foo, again, truncate_datetime="second") == {}
    assert 4 == 4


def test_content_hash() -> None:
    # the order of properties should not matter for the content hash
    g = MultiDiGraph()
    g.add_node("1", reported={"a": {"a": 1, "c": 2, "b": 3}, "c": 2, "b": 3, "d": "foo", "z": True, "kind": "a"})
    g.add_node("2", reported={"z": True, "c": 2, "b": 3, "a": {"b": 3, "c": 2, "a": 1}, "d": "foo", "kind": "a"})

    access = GraphAccess(g)
    sha1 = node(access, "1")["hash"]  # type: ignore
    sha2 = node(access, "2")["hash"]  # type: ignore
    assert sha1 == sha2


def test_root(graph_access: GraphAccess) -> None:
    assert graph_access.root() == "1"


def test_not_visited(graph_access: GraphAccess) -> None:
    graph_access.node(NodeId("1"))
    graph_access.node(NodeId("3"))
    not_visited = list(graph_access.not_visited_nodes())
    assert len(not_visited) == 2
    assert not_visited[0]["hash"] == "0546994698f57ed4a7a4463b0353397ebe4acc7196b9e9ca1dd7bc0b72245be6"
    assert not_visited[1]["hash"] == "0fead7dd5aff45049878e4c1b108b78c5fab594ac503eaba2199b71725241fce"


def test_edges(graph_access: GraphAccess) -> None:
    assert graph_access.has_edge("1", "2", EdgeTypes.default)
    assert not graph_access.has_edge("1", "9", EdgeTypes.default)
    assert graph_access.has_edge("2", "3", EdgeTypes.default)
    assert list(graph_access.not_visited_edges(EdgeTypes.default)) == [("1", "3"), ("2", "4"), ("3", "4")]
    assert list(graph_access.not_visited_edges(EdgeTypes.delete)) == [("1", "2"), ("1", "3"), ("1", "4")]


def test_desired(graph_access: GraphAccess) -> None:
    desired = {a["id"]: a["desired"] for a in graph_access.not_visited_nodes()}
    assert desired == {"1": {"name": "a"}, "2": {"name": "b"}, "3": {"name": "c"}, "4": {"name": "d"}}


def test_metadata(graph_access: GraphAccess) -> None:
    desired = {a["id"]: a["metadata"] for a in graph_access.not_visited_nodes()}
    assert desired == {"1": {"version": 1}, "2": {"version": 2}, "3": {"version": 3}, "4": {"version": 4}}


def test_flatten() -> None:
    js = {"id": "blub", "d": "2021-06-18T10:31:34Z", "i": 0, "s": "hello", "a": [{"a": "one"}, {"b": "two"}], "c": True}
    flat = GraphBuilder.flatten(js, AnyKind())
    assert flat == "blub 2021-06-18T10:31:34Z 0 hello one two"


def node(access: GraphAccess, node_id: NodeId) -> Optional[Json]:
    res = access.node(node_id)
    if res:
        return res
    else:
        raise AttributeError(f"Expected {node_id} to be defined!")


def test_builder(person_model: Model) -> None:
    max_m = {"id": "max", "kind": "Person", "name": "Max"}
    builder = GraphBuilder(person_model, uuid_str())
    builder.add_from_json({"id": "root", "reported": max_m})
    builder.add_from_json({"from": "root", "to": "2"})
    with pytest.raises(AssertionError) as no_node:
        builder.check_complete()
    assert str(no_node.value) == "2 was used in an edge definition but not provided as vertex!"
    builder.add_from_json({"id": "2", "reported": max_m})
    builder.add_from_json({"id": "3", "reported": max_m})
    with pytest.raises(AssertionError) as no_node:
        builder.check_complete()
    assert str(no_node.value) == "Given subgraph has more than one root: ['root', '3']"
    builder.add_from_json({"from": "root", "to": "3"})
    builder.check_complete()


def test_reassign_root(person_model: Model) -> None:
    max_m = {"id": "max", "kind": "Person", "name": "Max"}
    builder = GraphBuilder(person_model, uuid_str())
    builder.add_from_json({"id": "should_be_root", "reported": {"kind": "graph_root"}})
    builder.add_from_json({"id": "2", "reported": max_m})
    builder.add_from_json({"id": "3", "reported": max_m})
    builder.add_from_json({"from": "should_be_root", "to": "2"})
    builder.add_from_json({"from": "should_be_root", "to": "3"})
    builder.check_complete()
    access = GraphAccess(builder.graph)
    assert access.root() == "root"
    assert set(access.successors(NodeId("root"), EdgeTypes.default)) == {"2", "3"}


def test_replace_nodes(person_model: Model) -> None:
    builder = GraphBuilder(person_model, uuid_str())
    meta = {"metadata": {"replace": True}}
    builder.add_from_json({"id": "root", "reported": {"kind": "graph_root"}})
    builder.add_from_json({"id": "cloud", "reported": {"id": "cloud", "kind": "cloud"}, **meta})
    # also mark account and region as replace node -> the flags should be ignored!
    builder.add_from_json({"id": "account", "reported": {"id": "account", "kind": "account"}, **meta})
    builder.add_from_json({"id": "region", "reported": {"id": "region", "kind": "region"}, **meta})
    builder.add_from_json({"from": "root", "to": "cloud"})
    builder.add_from_json({"from": "cloud", "to": "account"})
    builder.add_from_json({"from": "account", "to": "region"})
    roots, _, gen = GraphAccess.merge_graphs(builder.graph)
    assert roots == ["cloud"]
    cloud, access = list(gen)[0]
    assert cloud == "cloud"
    assert set(access.nodes) == {"cloud", "account", "region"}


def multi_cloud_graph(replace_on: str) -> MultiDiGraph:
    g = MultiDiGraph()
    root = NodeId("root")

    def add_node(node_id: NodeId) -> None:
        kind = re.sub("_.*$", "", node_id)
        reported = {
            "id": f"id_{node_id}",
            "name": f"name_{node_id}",
            "kind": kind,
            "some": {"deep": {"nested": node_id}},
        }
        # for sake of testing: declare parent as phantom resource
        phantom = {"phantom": True} if kind == "parent" else {}
        g.add_node(
            node_id,
            id=node_id,
            replace=node_id.startswith(replace_on),
            reported=reported,
            metadata=phantom,
            kind=kind,
            kinds=[kind],
            kinds_set={kind},
        )

    def add_edge(from_node: str, to_node: str, edge_type: EdgeType = EdgeTypes.default) -> None:
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        g.add_edge(from_node, to_node, key, edge_type=edge_type)

    add_node(root)
    for cloud_d in ["aws", "gcp"]:
        cloud = NodeId(f"cloud_{cloud_d}")
        add_node(cloud)
        add_edge(root, cloud)
        for account_d in range(0, 3):
            account = NodeId(f"account_{cloud}_{account_d}")
            add_node(account)
            add_edge(cloud, account)
            add_edge(account, cloud, EdgeTypes.delete)
            for region_d in ["europe", "america", "asia", "africa", "antarctica", "australia"]:
                region = NodeId(f"region_{account}_{region_d}")
                add_node(region)
                add_edge(account, region)
                add_edge(region, account, EdgeTypes.delete)
                for parent_d in range(0, 3):
                    parent = NodeId(f"parent_{region}_{parent_d}")
                    add_node(parent)
                    add_edge(region, parent)
                    add_edge(parent, region, EdgeTypes.delete)
                    for children_d in range(0, 3):
                        children = NodeId(f"child_{parent}_{children_d}")
                        add_node(children)
                        add_edge(parent, children)
                        add_edge(children, parent, EdgeTypes.delete)

    return g


def cyclic_multi_graph(acyclic: bool) -> MultiDiGraph:
    g = MultiDiGraph()
    g.add_nodes_from([1, 2, 3])
    g.add_edge(1, 2, EdgeKey(1, 2, "default"))
    g.add_edge(1, 3, EdgeKey(1, 3, "default"))
    g.add_edge(2, 3, EdgeKey(2, 3, "default"))
    g.add_edge(2, 1, EdgeKey(2, 1, "delete"))
    g.add_edge(3, 2, EdgeKey(3, 2, "delete"))
    g.add_edge(3, 1, EdgeKey(3, 1, "delete"))
    if not acyclic:
        g.add_edge(3, 1, EdgeKey(3, 1, "default"))
        g.add_edge(1, 3, EdgeKey(1, 3, "delete"))
    return g


def test_sub_graphs_from_graph_cloud() -> None:
    graph = multi_cloud_graph("cloud")
    merges, parent, graph_it = GraphAccess.merge_graphs(graph)
    graphs = list(graph_it)
    assert len(graphs) == 2
    for root, succ in graphs:
        assert len(parent.nodes) == 3  # root + 2 x cloud
        assert succ.root().startswith("cloud")
        assert len(list(succ.not_visited_nodes())) == 237
        assert len(succ.nodes) == 238
        # make sure there is no node from another subgraph
        for node_id in succ.not_visited_nodes():
            assert succ.root() in node_id["id"]
        assert len(list(succ.not_visited_edges(EdgeTypes.default))) == 237
        assert len(list(succ.not_visited_edges(EdgeTypes.delete))) == 237


def test_sub_graphs_from_graph_account() -> None:
    graph = multi_cloud_graph("account")
    merges, parent, graph_it = GraphAccess.merge_graphs(graph)
    graphs = list(graph_it)
    assert len(graphs) == 6
    for root, succ in graphs:
        assert len(parent.nodes) == 9
        assert succ.root().startswith("account")
        assert len(list(succ.not_visited_nodes())) == 78
        assert len(succ.nodes) == 79
        # make sure there is no node from another subgraph
        for node_id in succ.not_visited_nodes():
            assert succ.root() in node_id["id"]
        assert len(list(succ.not_visited_edges(EdgeTypes.default))) == 78
        assert len(list(succ.not_visited_edges(EdgeTypes.delete))) == 78


def test_acyclic() -> None:
    assert GraphAccess(cyclic_multi_graph(acyclic=False)).is_acyclic_per_edge_type() is False
    assert GraphAccess(cyclic_multi_graph(acyclic=True)).is_acyclic_per_edge_type() is True


def test_predecessors() -> None:
    graph = GraphAccess(multi_cloud_graph("account"))
    child = NodeId("child_parent_region_account_cloud_gcp_2_europe_1_0")
    parent = NodeId("parent_region_account_cloud_gcp_2_europe_1")
    region = NodeId("region_account_cloud_gcp_2_europe")

    # default: region -> parent -> child
    assert list(graph.predecessors(child, EdgeTypes.default)) == [parent]
    assert list(graph.predecessors(parent, EdgeTypes.default)) == [region]
    assert child in list(graph.successors(parent, EdgeTypes.default))
    assert parent in list(graph.successors(region, EdgeTypes.default))

    # delete: child -> parent -> region
    assert list(graph.successors(child, EdgeTypes.delete)) == [parent]
    assert list(graph.successors(parent, EdgeTypes.delete)) == [region]
    assert parent in list(graph.successors(child, EdgeTypes.delete))
    assert region in list(graph.successors(parent, EdgeTypes.delete))


def test_ancestor_of() -> None:
    nid1 = NodeId("child_parent_region_account_cloud_gcp_1_europe_1_0")
    acc1 = "account_cloud_gcp_1"
    acc2 = "account_cloud_gcp_2"
    g = multi_cloud_graph("account")

    graph = GraphAccess(g)
    assert graph.ancestor_of(nid1, EdgeTypes.default, "root") is not None
    assert graph.ancestor_of(nid1, EdgeTypes.delete, "root") is None
    assert graph.ancestor_of(nid1, EdgeTypes.default, "foo") is None
    assert graph.ancestor_of(nid1, EdgeTypes.default, "foo") is None
    assert graph.ancestor_of(nid1, EdgeTypes.default, "account")["id"] == acc1  # type: ignore

    # add another "shorter" edge from acc2 -> nid1, so it is shorter that from acc1 -> nid1
    key = GraphAccess.edge_key(acc2, nid1, EdgeTypes.default)
    g.add_edge(acc2, nid1, key, edge_type=EdgeTypes.default)
    assert graph.ancestor_of(nid1, EdgeTypes.default, "account")["id"] == acc2  # type: ignore


def test_resolve_graph_data() -> None:
    g = multi_cloud_graph("account")
    graph = GraphAccess(g)
    graph.resolve()

    # ancestor data should be stored in metadata
    n1 = AccessJson(graph.node("child_parent_region_account_cloud_gcp_1_europe_1_0"))  # type: ignore
    assert n1.refs.region_id == "region_account_cloud_gcp_1_europe"
    assert n1.ancestors.account.reported.id == "id_account_cloud_gcp_1"
    assert n1.ancestors.account.reported.name == "name_account_cloud_gcp_1"
    assert n1.ancestors.region.reported.id == "id_region_account_cloud_gcp_1_europe"
    assert n1.ancestors.region.reported.name == "name_region_account_cloud_gcp_1_europe"
    # make sure there is no summary
    assert n1.descendant_summary == AccessNone(None)

    r1 = AccessJson(graph.node("region_account_cloud_gcp_1_europe"))  # type: ignore
    assert r1.metadata.descendant_summary == {"child": 9}
    assert r1.metadata.descendant_count == 9
    r2 = AccessJson(graph.node("account_cloud_gcp_1"))  # type: ignore
    assert r2.metadata.descendant_summary == {"child": 54, "region": 6}
    assert r2.metadata.descendant_count == 60
    r3 = AccessJson(graph.node("cloud_gcp"))  # type: ignore
    assert r3.metadata.descendant_summary == {"child": 162, "region": 18, "account": 3}
    assert r3.metadata.descendant_count == 183


def test_model_size(person_model: Model) -> None:
    builder = GraphBuilder(person_model, uuid_str())
    tags1 = {"foo": "bar", "bla": "blub" * 22}
    a1 = {"kind": "Address", "id": "a1", "zip": "s1", "city": "c1", "list": ["ccc"]}
    a2 = {"kind": "Address", "id": "aa2", "zip": "s2", "city": "gotham", "tags": tags1}
    p1 = dict(kind="Person", id="pp1", name="pp1", address=a1, list=["a", "bb"], tags=tags1)
    p2 = dict(kind="Person", id="p2", name="ppp2", addresses=[a1, a2], other_addresses=dict(home=a1, work=a2))
    builder.add_node(NodeId("p1"), p1)
    builder.add_node(NodeId("p2"), p2)

    def kind_props(p: ComplexKind) -> Dict[str, int]:
        return {p.name: p.meta_get("len", int, 0) for p in p.all_props() if p.meta("len", int)}

    base = {"id": 3, "kind": 7, "list": 3, "tags": 88}  # base properties shared by person and address
    assert kind_props(cast(ComplexKind, person_model["Person"])) == base | {"name": 4}
    assert kind_props(cast(ComplexKind, person_model["Address"])) == base | {"city": 6, "zip": 2}



from typing import List, Optional, Callable, Dict, Any, Awaitable, cast
from pathlib import Path
import yaml
import logging
from datetime import timedelta

from resotocore.core_config import config_model
from resotocore.model.typed_model import from_js
from resotocore.service import Service
from resotocore.types import Json
from resotocore.ids import ConfigId
from resotocore.util import Periodic
from resotocore.config import ConfigOverride
from resotocore.db.modeldb import ModelDb
from resotocore.model.model import Model, Kind
from resotolib.utils import merge_json_elements
from deepdiff import DeepDiff
import aiofiles.os as aos
import aiofiles
import jsons

log = logging.getLogger("config_override_service")


async def model_from_db(model_db: ModelDb) -> Model:
    kinds = [kind async for kind in model_db.all()]
    return Model.from_kinds(list(kinds))


class ConfigOverrideService(ConfigOverride, Service):
    def __init__(
        self, override_paths: List[Path], get_configs_model: Callable[[], Awaitable[Model]], sleep_time: float = 10.0
    ):
        self.override_paths = override_paths
        self._get_configs_model = get_configs_model

        self.overrides: Dict[ConfigId, Json] = {}
        self.override_change_hooks: List[Callable[[Dict[ConfigId, Json]], Awaitable[Any]]] = []
        self.watcher: Periodic = Periodic(
            "config_overrides_watcher", self.check_config_changes, timedelta(seconds=sleep_time)
        )
        self.mtime_hash: int = 0

    async def load(self) -> None:
        self.overrides = await self._get_overrides()

    def add_override_change_hook(self, hook: Callable[[Dict[ConfigId, Json]], Awaitable[Any]]) -> None:
        self.override_change_hooks.append(hook)

    def coerce_and_validate(self, config: Json, model: Model) -> Json:
        final_config = {}
        for key, value in config.items():
            if key in model:
                try:
                    value_kind = model[key]
                    coerced = value_kind.check_valid(value, normalize=False, config_context=True)
                    final_config[key] = value_kind.sort_json(coerced or value)
                except Exception as ex:
                    raise AttributeError(f"Error validating section {key}: {ex}") from ex
            else:
                final_config[key] = value

        return final_config

    async def _get_overrides(self, silent: bool = False) -> Dict[ConfigId, Json]:
        if not self.override_paths:
            return {}

        # all config files that will be used
        config_files: List[Path] = []
        # collect them all
        for path in self.override_paths:
            if path.is_dir():
                config_files.extend(
                    [file for file in path.iterdir() if file.is_file() and file.suffix in (".yml", ".yaml", ".json")]
                )
            elif path.suffix in (".yml", ".yaml", ".json"):
                config_files.append(path)
            else:
                log.warning(f"Config override path {path} is neither a directory nor a yaml/json file, skipping.")

        # json with all merged overrides for all components such as resotocore, resotoworker, etc.
        overrides_json: Json = {}

        model = await self._get_configs_model()

        # merge all provided overrides into a single object, preferring the values from the last override
        for config_file in config_files:
            async with aiofiles.open(config_file) as f:
                try:

                    def is_yaml(path: Path) -> bool:
                        return path.suffix in (".yml", ".yaml")

                    content = await f.read()
                    raw_yaml = yaml.safe_load(content) if is_yaml(config_file) else jsons.loads(content, Json)
                    validated = self.coerce_and_validate(raw_yaml, model)
                    with_config_id = {config_file.stem: validated}
                    merged = merge_json_elements(overrides_json, with_config_id)
                    overrides_json = cast(Json, merged)
                except Exception as e:
                    log.warning(f"Can't load the config override {config_file}, skipping. Reason: {e}")

        def is_dict(config_id: str, obj: Any) -> bool:
            if not isinstance(obj, dict):
                if not silent:
                    log.warning(f"Config override with id {config_id} contains invalid data, skipping.")
                return False
            return True

        # dict with all overrides for all config ids, such as resoto.core, resoto.worker, etc.
        all_config_overrides: Dict[ConfigId, Json] = {
            ConfigId(k): v for k, v in overrides_json.items() if is_dict(k, v)
        }

        if all_config_overrides is None:
            if not silent:
                log.warning("No config overrides found.")
            return {}

        return all_config_overrides

    def get_override(self, config_id: ConfigId) -> Optional[Json]:
        return self.overrides.get(config_id)

    def get_all_overrides(self) -> Dict[ConfigId, Json]:
        return self.overrides

    async def check_config_changes(self) -> None:
        # all config files that needs to be checked for changes
        config_files: List[Path] = []
        # do a flatmap on directories
        for path in self.override_paths:
            if await aos.path.isdir(path):
                config_files.extend(
                    [
                        Path(entry.path)
                        for entry in await aos.scandir(path)  # scandir avoids extra syscalls
                        if entry.is_file() and Path(entry.path).suffix in (".yml", ".yaml")
                    ]
                )
            else:
                config_files.append(path)

        # a quick optimization to avoid reading the files if none has been changed
        mtime_hash = 0
        config_files = sorted(config_files)
        for file in config_files:
            mtime_hash = hash((mtime_hash, (await aos.stat(file)).st_mtime))

        if mtime_hash == self.mtime_hash:
            return
        self.mtime_hash = mtime_hash

        overrides = await self._get_overrides(silent=True)
        diff = DeepDiff(self.overrides, overrides, ignore_order=True)

        if diff:
            self.overrides = overrides
            for hook in self.override_change_hooks:
                await hook(self.overrides)

    async def start(self) -> None:
        await self.load()
        await self.watcher.start()

    async def stop(self) -> None:
        await self.watcher.stop()


async def override_config_for_startup(override_path: List[Path]) -> ConfigOverride:
    """
    Minimal version that is used only for bootstrapping the core
    """

    async def core_config_model() -> Model:
        return Model.from_kinds(from_js(config_model(), List[Kind]))

    config_override_service = ConfigOverrideService(override_path, core_config_model)
    await config_override_service.load()
    return config_override_service



from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
import os
from attrs import evolve
from pytest import fixture

from resotocore import core_config
from resotocore.core_config import (
    parse_config,
    CoreConfig,
    ApiConfig,
    config_model,
    EditableConfig,
    WorkflowConfig,
    migrate_core_config,
    migrate_command_config,
    CustomCommandsConfig,
    alias_templates,
    ResotoCoreCommandsRoot,
    ResotoCoreConfigId,
)
from resotocore.system_start import parse_args
from resotocore.model.typed_model import to_js, from_js
from resotocore.types import Json
from resotocore.util import value_in_path


def test_parse_empty(default_config: CoreConfig) -> None:
    result = parse_config(parse_args(["--analytics-opt-out"]), {}, lambda: None)
    assert result == default_config


def test_parse_broken(config_json: Json) -> None:
    # config_json is a valid parsable config
    cfg = deepcopy(config_json)

    # adjust the config: rename web_hosts -> hosts, and https_port -> port
    hosts = cfg["resotocore"]["api"]["web_hosts"]
    port = cfg["resotocore"]["api"]["https_port"]
    cfg["resotocore"]["api"]["hosts"] = hosts
    cfg["resotocore"]["api"]["port"] = port
    del cfg["resotocore"]["api"]["web_hosts"]
    del cfg["resotocore"]["api"]["https_port"]

    # parse this configuration
    parsed = parse_config(parse_args(["--analytics-opt-out"]), cfg, lambda: None)
    parsed_json = to_js(parsed.editable, strip_attr="kind")

    # web_hosts and https_port were not available and are reverted to the default values
    default = EditableConfig()
    assert parsed.api.web_hosts != hosts
    assert parsed.api.web_hosts == default.api.web_hosts
    assert parsed.api.https_port != port
    assert parsed.api.https_port == default.api.https_port

    # other config values are still unchanged
    assert parsed_json["cli"] == config_json["resotocore"]["cli"]
    assert parsed_json["runtime"] == config_json["resotocore"]["runtime"]
    assert parsed_json["graph_update"] == config_json["resotocore"]["graph_update"]


def test_read_config(config_json: Json) -> None:
    parsed = parse_config(parse_args(["--analytics-opt-out"]), config_json, lambda: None)
    assert parsed.json() == config_json


def test_override_via_cmd_line(default_config: CoreConfig) -> None:
    config = {"runtime": {"debug": False}}
    parsed = parse_config(parse_args(["--debug"]), config, lambda: None)
    assert parsed.runtime.debug == True


# noinspection PyTypeChecker
def test_validate() -> None:
    assert EditableConfig().validate() is None
    assert EditableConfig(api=ApiConfig(tsdb_proxy_url="wrong")).validate() == {
        "api": [{"tsdb_proxy_url": ["url is missing host", "url is missing scheme"]}]
    }
    assert EditableConfig(api=ApiConfig(max_request_size=1)).validate() == {
        "api": [{"max_request_size": ["min value is 1048576"]}]
    }
    assert EditableConfig(api=ApiConfig(tsdb_proxy_url="wrong", max_request_size=1)).validate() == {
        "api": [
            {
                "tsdb_proxy_url": ["url is missing host", "url is missing scheme"],
                "max_request_size": ["min value is 1048576"],
            }
        ]
    }
    assert EditableConfig(workflows={"foo": WorkflowConfig("bla")}).validate() == {
        "workflows": [{"foo": [{"schedule": ["Invalid cron expression: Wrong number of fields; got 1, expected 5"]}]}]
    }


def test_config_override(config_json: Json) -> None:
    # config_json is a valid parsable config
    cfg = deepcopy(config_json)

    overrides = {"resoto.core": {"resotocore": {"api": {"web_hosts": ["11.12.13.14"], "https_port": "$(WEB_PORT)"}}}}

    os.environ["WEB_PORT"] = "1337"

    # parse this configuration
    parsed = parse_config(
        parse_args(["--analytics-opt-out"]),
        cfg,
        lambda: overrides.get(ResotoCoreConfigId),
    )
    assert parsed.api.web_hosts == ["11.12.13.14"]
    assert parsed.api.https_port == 1337


def test_model() -> None:
    model = config_model()
    assert {m["fqn"] for m in model} == {
        "custom_commands",
        "resotocore",
        "resotocore_api_config",
        "resotocore_certificate_config",
        "resotocore_cli_alias_template",
        "resotocore_cli_alias_template_parameter",
        "resotocore_cli_config",
        "resotocore_graph_update_config",
        "resotocore_runtime_config",
        "resotocore_workflow_config",
    }


def test_in_docker() -> None:
    with TemporaryDirectory() as tmp:
        path = Path(tmp, "git.hash")
        path.write_text("foo", encoding="utf-8")
        stored = core_config.GitHashFile
        core_config.GitHashFile = str(path)
        assert core_config.inside_docker() is True
        assert core_config.git_hash_from_file() == "foo"
        assert core_config.default_hosts() == ["0.0.0.0"]
        core_config.GitHashFile = "/this/path/does/not/exist"
        assert core_config.inside_docker() is False
        assert core_config.git_hash_from_file() is None
        assert core_config.default_hosts() == ["localhost"]
        core_config.GitHashFile = stored


def test_migration() -> None:
    cfg1 = migrate_core_config(dict(resotocore=dict(runtime=dict(analytics_opt_out=True))))
    assert value_in_path(cfg1, "resotocore.runtime.usage_metrics") is False
    assert value_in_path(cfg1, "resotocore.runtime.analytics_opt_out") is None
    cfg2 = migrate_core_config(dict(resotocore=dict(runtime=dict(usage_metrics=True))))
    assert value_in_path(cfg2, "resotocore.runtime.usage_metrics") is True
    assert value_in_path(cfg1, "resotocore.runtime.analytics_opt_out") is None
    cfg3 = migrate_core_config(dict(resotocore=dict(runtime=dict(analytics_opt_out=True, usage_metrics=True))))
    assert value_in_path(cfg3, "resotocore.runtime.usage_metrics") is True
    assert value_in_path(cfg1, "resotocore.runtime.analytics_opt_out") is None
    cfg4 = migrate_core_config(dict(resotocore=dict(api=dict(web_port=1234))))
    assert value_in_path(cfg4, "resotocore.api.https_port") == 1234
    assert value_in_path(cfg4, "resotocore.api.web_port") is None


def test_migrate_commands() -> None:
    # default configuration does not need migration
    assert migrate_command_config(CustomCommandsConfig().json()) is None
    # an empty configuration is migrated to the default configuration
    assert migrate_command_config(CustomCommandsConfig(commands=[]).json()) == CustomCommandsConfig().json()
    # an existing configuration is not destroyed
    example = evolve(alias_templates()[0], name="my-test-cmd")
    custom = CustomCommandsConfig(commands=[example])
    migrated_json: Json = migrate_command_config(custom.json())  # type: ignore
    assert migrated_json != custom.json()
    migrated = from_js(migrated_json.get(ResotoCoreCommandsRoot), CustomCommandsConfig)
    assert any(cmd == example for cmd in migrated.commands)
    assert len(migrated.commands) == len(alias_templates()) + 1


@fixture
def config_json() -> Json:
    return {
        "resotocore": {
            "api": {
                "access_token_expiration_seconds": 3600,
                "web_hosts": ["1.2.3.4"],
                "https_port": 443,
                "http_port": 80,
                "web_path": "/",
                "tsdb_proxy_url": "test",
                "max_request_size": 5242880,
                "host_certificate": {
                    "common_name": "test",
                    "san_dns_names": ["test.example.com"],
                    "san_ip_addresses": ["4.3.2.1"],
                    "include_loopback": False,
                },
            },
            "cli": {
                "default_graph": "foo",
                "default_section": "bla",
                "alias_templates": [
                    {
                        "info": "Test command",
                        "name": "test",
                        "parameters": [{"name": "test", "default": "test argument", "description": "easy"}],
                        "template": "do something",
                    }
                ],
            },
            "graph_update": {
                "abort_after_seconds": 1234,
                "merge_max_wait_time_seconds": 4321,
                "keep_history": True,
                "parallel_imports": 5,
            },
            "runtime": {
                "usage_metrics": False,
                "debug": True,
                "log_level": "WARN",
                "plantuml_server": "https://foo",
                "start_collect_on_subscriber_connect": True,
            },
            "workflows": {
                "collect_and_cleanup": {
                    "schedule": "0 0 0 0 *",
                }
            },
        }
    }



from datetime import timedelta
from typing import Any, AsyncIterator

from deepdiff import DeepDiff
from pytest import fixture, mark

from resotocore.db.subscriberdb import SubscriberDb
from resotocore.message_bus import MessageBus
from resotocore.model.typed_model import to_js, from_js
from resotocore.task.model import Subscription, Subscriber
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.ids import SubscriberId

from tests.resotocore.db.entitydb import InMemoryDb


@fixture
def in_mem_db() -> SubscriberDb:
    return InMemoryDb[SubscriberId, Subscriber](Subscriber, lambda x: x.id)


@fixture
async def handler(in_mem_db: SubscriberDb) -> AsyncIterator[SubscriptionHandler]:
    async with SubscriptionHandler(in_mem_db, MessageBus()) as handler:
        await handler.add_subscription(SubscriberId("sub_1"), "test", True, timedelta(seconds=3))
        yield handler


def test_json_marshalling_subscription() -> None:
    roundtrip(Subscription("test"))
    roundtrip(Subscription("test", False, timedelta(days=1)))
    assert from_js({"message_type": "foo"}, Subscription) == Subscription("foo")
    assert from_js({"message_type": "foo"}, Subscription) == Subscription("foo")
    assert from_js({"message_type": "a", "timeout": 86400}, Subscription) == Subscription("a", True, timedelta(days=1))


def test_json_marshalling_subscribers() -> None:
    subscriptions = [
        Subscription("a", True, timedelta(seconds=1)),
        Subscription("b", True, timedelta(minutes=1)),
        Subscription("c", True, timedelta(hours=1)),
        Subscription("d", False, timedelta(days=1)),
        Subscription("e", False, timedelta(weeks=1)),
    ]
    roundtrip(Subscriber.from_list(SubscriberId("foo"), []))
    roundtrip(Subscriber.from_list(SubscriberId("foo"), subscriptions))


@mark.asyncio
async def test_subscribe(handler: SubscriptionHandler, in_mem_db: SubscriberDb) -> None:
    # register first time
    result = await handler.add_subscription(SubscriberId("foo"), "event_bla", True, timedelta(seconds=3))
    assert len(result.subscriptions) == 1
    assert result.subscriptions["event_bla"].message_type == "event_bla"
    # should be persisted in database as well
    assert len((await in_mem_db.get("foo")).subscriptions) == 1  # type: ignore
    # register again is ignored
    result = await handler.add_subscription(SubscriberId("foo"), "event_bla", True, timedelta(seconds=3))
    assert len(result.subscriptions) == 1
    assert result.subscriptions["event_bla"].message_type == "event_bla"
    # should be persisted in database as well
    assert len((await in_mem_db.get("foo")).subscriptions) == 1  # type: ignore


@mark.asyncio
async def test_unsubscribe(handler: SubscriptionHandler, in_mem_db: SubscriberDb) -> None:
    # register first time
    subscriber_id = SubscriberId("foo")
    subs = [Subscription("event_bla"), Subscription("event_bar")]
    result = await handler.update_subscriptions(subscriber_id, subs)
    assert len(result.subscriptions) == 2
    updated = await handler.remove_subscription(subscriber_id, "event_bla")
    assert len(updated.subscriptions) == 1
    # should be persisted in database as well
    assert len((await in_mem_db.get(subscriber_id)).subscriptions) == 1  # type: ignore
    # second time should be ignored
    updated = await handler.remove_subscription(subscriber_id, "event_bla")
    assert len(updated.subscriptions) == 1
    # last subscription is removed
    updated = await handler.remove_subscription(subscriber_id, "event_bar")
    assert len(updated.subscriptions) == 0
    # should be persisted in database as well
    assert await in_mem_db.get(subscriber_id) is None


@mark.asyncio
async def test_get_subscriber(handler: SubscriptionHandler) -> None:
    result = await handler.get_subscriber(SubscriberId("sub_1"))
    assert result
    assert result.id == "sub_1"


@mark.asyncio
async def test_by_event_type(handler: SubscriptionHandler) -> None:
    result = await handler.list_subscriber_for("test")
    assert len(result) == 1
    assert result[0].subscriptions["test"].message_type == "test"
    result2 = await handler.list_subscriber_for("does not exist")
    assert len(result2) == 0


def roundtrip(obj: Any) -> None:
    js = to_js(obj)
    again = from_js(js, type(obj))
    assert DeepDiff(obj, again) == {}, f"Json: {js} serialized as {again}"



import base64
import hashlib
import secrets
from typing import Optional, Dict, List, Callable, TypeVar, Awaitable

from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.config import ConfigHandler, ConfigEntity
from resotocore.db.db_access import DbAccess
from resotocore.ids import Email, Password
from resotocore.model.typed_model import from_js, to_js
from resotocore.types import Json
from resotocore.user import UserManagement, UsersConfigId, ResotoUser, UsersConfigRoot
from resotocore.util import value_in_path_get, value_in_path

ALGORITHM = "pbkdf2_sha512"
DELIMITER = "$"
ITERATIONS = 210000  # https://cheatsheetseries.owasp.org/cheatsheets/Password_Storage_Cheat_Sheet.html#pbkdf2

T = TypeVar("T")


class UserManagementService(UserManagement):
    def __init__(self, db_access: DbAccess, config_handler: ConfigHandler, event_sender: AnalyticsEventSender):
        self.db_access = db_access
        self.config_handler = config_handler
        self.event_sender = event_sender

    @staticmethod
    def hash_password(password: Password, salt: Optional[str] = None) -> str:
        if salt is None:
            salt = secrets.token_hex(16)
        pw_hash = hashlib.pbkdf2_hmac("sha512", password.encode("utf-8"), salt.encode("utf-8"), ITERATIONS)
        b64_hash = base64.b64encode(pw_hash).decode("ascii").strip()
        return f"{DELIMITER}{ALGORITHM}{DELIMITER}{salt}{DELIMITER}{b64_hash}"

    @staticmethod
    def verify_password(password: Password, password_hash: str) -> bool:
        if (password_hash or "").count(DELIMITER) != 3:
            return False
        _, algorithm, salt, _ = password_hash.split(DELIMITER)
        assert algorithm == ALGORITHM
        compare_hash = UserManagementService.hash_password(password, salt)
        return secrets.compare_digest(password_hash, compare_hash)

    async def has_users(self) -> bool:
        user_config = await self.config_handler.get_config(UsersConfigId)
        return bool(value_in_path(user_config.config, [UsersConfigRoot, "users"])) if user_config else False

    async def create_first_user(self, company: str, fullname: str, email: Email, password: Password) -> ResotoUser:
        assert not await self.has_users()  # only allowed if no users exist
        await self.db_access.system_data_db.update_info(company=company)
        hashed = self.hash_password(password)
        user = ResotoUser(fullname=fullname, password_hash=hashed, roles={"admin"})
        await self.config_handler.put_config(
            ConfigEntity(UsersConfigId, {UsersConfigRoot: {"users": {email: to_js(user)}}})
        )
        await self.event_sender.core_event(
            CoreEvent.FirstUserCreated, {"company": company, "fullname": fullname, "email": email}
        )
        return user

    async def login(self, email: Email, password: Password) -> Optional[ResotoUser]:
        user_config = await self.config_handler.get_config(UsersConfigId)
        if user_config:
            users: Dict[Email, Json] = value_in_path_get(user_config.config, [UsersConfigRoot, "users"], {})
            if (user := users.get(email)) and self.verify_password(password, user.get("password_hash", "")):
                return from_js(user, ResotoUser)
        return None

    async def create_user(self, email: Email, fullname: str, password: Password, roles: List[str]) -> ResotoUser:
        async def fn(users: Dict[Email, ResotoUser]) -> ResotoUser:
            if email in users:
                raise ValueError(f"User with email {email} already exists")
            if not email or email.startswith("@") or email.endswith("@") or email.count("@") != 1:
                raise ValueError(f"Invalid email address {email}")
            hashed = self.hash_password(password)
            user = ResotoUser(fullname, password_hash=hashed, roles=set(roles))
            users[email] = user
            await self.event_sender.core_event(CoreEvent.UserCreated, {"fullname": fullname, "email": email})
            return user

        return await self.__change_users(fn)

    async def delete_user(self, email: Email) -> Optional[ResotoUser]:
        async def fn(users: Dict[Email, ResotoUser]) -> Optional[ResotoUser]:
            return users.pop(email, None)

        return await self.__change_users(fn)

    async def update_user(
        self, email: Email, *, password: Optional[Password] = None, roles: Optional[List[str]] = None
    ) -> ResotoUser:
        async def fn(users: Dict[Email, ResotoUser]) -> ResotoUser:
            if email not in users:
                raise ValueError(f"User with email {email} does not exist")
            user = users[email]
            if password is not None:
                user.password_hash = self.hash_password(password)
            if roles is not None:
                user.roles = set(roles)
            users[email] = user
            return user

        return await self.__change_users(fn)

    async def __change_users(self, fn: Callable[[Dict[Email, ResotoUser]], Awaitable[T]]) -> T:
        users = await self.users()
        result = await fn(users)
        users_raw = {email: to_js(user) for email, user in users.items()}
        await self.config_handler.put_config(ConfigEntity(UsersConfigId, {UsersConfigRoot: {"users": users_raw}}))
        return result

    async def user(self, email: Email) -> Optional[ResotoUser]:
        users = await self.users()
        return users.get(email)

    async def users(self) -> Dict[Email, ResotoUser]:
        user_config = await self.config_handler.get_config(UsersConfigId)
        users_raw: Json = value_in_path_get(user_config.config, [UsersConfigRoot, "users"], {}) if user_config else {}
        return {Email(email): from_js(data, ResotoUser) for email, data in users_raw.items()}



from resotocore.validator import Validator


def test_path_exist() -> None:
    v = Validator()
    assert v.validate({"path": "does not exist"}, {"path": {"path_exists": True}}) is False
    assert v.errors == {"path": ["Path does not exist: does not exist"]}
    v = Validator()
    assert v.validate({"path": "/"}, {"path": {"path_exists": True}}) is True


def test_is_url() -> None:
    v = Validator()
    assert v.validate({"url": "does not exist"}, {"url": {"is_url": True}}) is False
    assert v.errors == {"url": ["url is missing host", "url is missing scheme"]}
    v = Validator()
    assert v.validate({"url": "https://resoto.com"}, {"url": {"is_url": True}}) is True


def test_is_cron() -> None:
    v = Validator()
    assert v.validate({"cron": "* * * * *"}, {"cron": {"is_cron": True}}) is True
    assert v.validate({"cron": "no cron expression"}, {"cron": {"is_cron": True}}) is False



import asyncio
from datetime import timedelta

import pytest

from resotocore.analytics import InMemoryEventSender, CoreEvent
from resotocore.analytics.recurrent_events import emit_recurrent_events
from resotocore.message_bus import MessageBus
from resotocore.task.model import Subscriber
from resotocore.worker_task_queue import WorkerTaskQueue
from tests.resotocore.db.entitydb import InMemoryDb
from resotocore.model.model import Model
from resotocore.task.subscribers import SubscriptionHandler
from tests.resotocore.model import ModelHandlerStatic


@pytest.mark.asyncio
async def test_emit_recurrent_events() -> None:
    message_bus = MessageBus()
    sender = InMemoryEventSender()
    model = ModelHandlerStatic(Model.empty())
    sub = SubscriptionHandler(InMemoryDb(Subscriber, lambda x: x.id), message_bus)
    queue = WorkerTaskQueue()
    fast = timedelta(seconds=0.001)
    periodic = emit_recurrent_events(sender, model, sub, queue, message_bus, fast, fast)
    await periodic.start()
    while len(sender.events) < 3:
        await asyncio.sleep(0.01)
    await periodic.stop()
    model_event, subscriber_event, worker_event = sender.events[0:3]
    assert model_event.kind == CoreEvent.ModelInfo
    assert model_event.counters["model_count"] == 0
    assert subscriber_event.kind == CoreEvent.SubscriberInfo
    assert subscriber_event.counters["subscriber_count"] == 0
    assert worker_event.kind == CoreEvent.WorkerQueueInfo
    assert worker_event.counters["worker_count"] == 0



from typing import List, Optional, AsyncIterator, cast, Tuple
from asyncio import Lock

import logging

from resotocore.db.db_access import DbAccess
from resotocore.db.graphdb import EventGraphDB
from resotocore.util import utc_str, UTC_Date_Format_short, utc
from resotocore.ids import GraphName, TaskDescriptorId
from resotocore.service import Service
from resotocore.util import check_graph_name, Periodic
from resotocore.types import Json
from resotocore.model.model import Kind
from resotocore.model.typed_model import from_js, to_js_str
from resotocore.model.graph_access import EdgeTypes
from resotocore.db.async_arangodb import AsyncCursor
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.core_config import SnapshotsScheduleConfig, ResotoCoreSnapshotsConfigId
from resotocore.task import TaskHandler
from resotocore.task.task_description import Job, ExecuteCommand, TimeTrigger
from json import loads, dumps
from datetime import datetime, timedelta
from dateutil.parser import parse
from attrs import frozen
import re


log = logging.getLogger(__name__)


class GraphManager(Service):
    def __init__(
        self,
        db_access: DbAccess,
        default_snapshots_config: SnapshotsScheduleConfig,
        config_handler: CoreConfigHandler,
        task_handler: TaskHandler,
    ) -> None:
        self.db_access = db_access
        self.lock: Optional[Lock] = None
        self.task_handler = task_handler
        self.default_snapshots_config = default_snapshots_config
        self.config_handler = config_handler
        self.snapshot_cleanup_worker: Optional[Periodic] = None

    async def __setup_cleanup_old_snapshots_worker(self, snapshots_config: SnapshotsScheduleConfig) -> None:
        if self.snapshot_cleanup_worker:
            await self.snapshot_cleanup_worker.stop()

        self.snapshot_cleanup_worker = Periodic(
            "snapshot_cleanup_worker", lambda: self._clean_outdated_snapshots(snapshots_config), timedelta(seconds=60)
        )
        await self.snapshot_cleanup_worker.start()

    async def _clean_outdated_snapshots(self, snapshots_config: SnapshotsScheduleConfig) -> None:
        # get all existing snapshots
        existing_snapshots = await self.list("snapshot-.*")

        snapshots_to_keep: List[Tuple[str, int]] = []
        for label, schedule in snapshots_config.snapshots.items():
            regex = rf"snapshot-\w+-{label}-.*"
            snapshots_to_keep.append((regex, schedule.retain))

        # delete all snapshots that are outdated
        for regex, retain in snapshots_to_keep:
            snapshots = [snapshot for snapshot in existing_snapshots if re.match(regex, snapshot)]
            snapshots.sort(reverse=True)
            for snapshot in snapshots[retain:]:
                await self.delete(snapshot)

    async def _on_config_updated(self, config_id: str) -> None:
        if config_id == ResotoCoreSnapshotsConfigId:
            job_prefix = "resoto:snapshots:"
            # get the new config or use the default
            snapshots_config = SnapshotsScheduleConfig()
            try:
                new_config = await self.config_handler.config_handler.get_config(ResotoCoreSnapshotsConfigId)
                if new_config:
                    snapshots_config = from_js(new_config.config, SnapshotsScheduleConfig)
            except Exception as e:
                log.error(f"Can not parse snapshot schedule. Fall back to defaults. Reason: {e}", exc_info=e)

            # recreate the cleanup worker according to the new schedule
            await self.__setup_cleanup_old_snapshots_worker(snapshots_config)

            # cancel all existing snapshot jobs
            existing_jobs = [job for job in await self.task_handler.list_jobs() if job.id.startswith(job_prefix)]
            for job in existing_jobs:
                await self.task_handler.delete_job(job.id, force=True)

            # schedule new snapshot jobs for the current graph
            for label, schedule in snapshots_config.snapshots.items():
                job = Job(
                    uid=TaskDescriptorId(f"{job_prefix}{label}"),
                    command=ExecuteCommand(f"graph snapshot {label}"),
                    timeout=timedelta(minutes=5),
                    trigger=TimeTrigger(schedule.schedule),
                )

                await self.task_handler.add_job(job, force=True)

    async def start(self) -> None:
        self.lock = Lock()
        # initialize the snapshot schedule
        await self._on_config_updated(ResotoCoreSnapshotsConfigId)
        # subscribe to config updates to update the snapshot schedule
        self.config_handler.add_callback(self._on_config_updated)
        await self.__setup_cleanup_old_snapshots_worker(self.default_snapshots_config)

    async def stop(self) -> None:
        if self.snapshot_cleanup_worker:
            await self.snapshot_cleanup_worker.stop()
            self.snapshot_cleanup_worker = None

    async def list(self, pattern: Optional[str]) -> List[GraphName]:
        return [key for key in await self.db_access.list_graphs() if pattern is None or re.match(pattern, key)]

    async def snapshot_at(self, *, time: datetime, graph_name: GraphName) -> Optional[GraphName]:
        regex = rf"snapshot-{graph_name}-.*-(.+)"
        graphs = await self.list(regex)
        graphs_with_time = []
        for graph in graphs:
            match = re.match(regex, graph)
            if match:
                graphs_with_time.append((parse(match.group(1)), graph))

        graphs_with_time.sort(reverse=True, key=lambda x: x[0])
        # take the first graph that is older than the given time
        for graph_time, graph in graphs_with_time:
            if graph_time <= time:
                return graph

        # nothing found
        return None

    async def copy(
        self,
        source: GraphName,
        destination: GraphName,
        replace_existing: bool,
        validate_name: bool = True,
        to_snapshot: bool = False,
    ) -> GraphName:
        if not self.lock:
            raise RuntimeError("GraphManager has not been started")

        async with self.lock:
            if not await self.db_access.db.has_graph(source):
                raise ValueError(f"Source graph {source} does not exist")

            if await self.db_access.db.has_graph(destination):
                if replace_existing:
                    await self.delete(destination)
                else:
                    raise ValueError(f"Destination graph {destination} already exists")
            return await self._copy_graph(source, destination, validate_name, to_snapshot)

    async def _copy_graph(
        self, source: GraphName, destination: GraphName, validate_name: bool = True, to_snapshot: bool = False
    ) -> GraphName:
        destination = GraphName(_compress_timestamps(destination))

        if validate_name:
            check_graph_name(destination)

        if not await self.db_access.db.has_graph(source):
            raise ValueError(f"Source graph {source} does not exist")

        source_db = self.db_access.get_graph_db(source, no_check=True)

        await source_db.copy_graph(destination, to_snapshot)

        source_model_db = await self.db_access.get_graph_model_db(source)
        destination_model_db = await self.db_access.get_graph_model_db(destination)

        model_kinds = [kind async for kind in source_model_db.all()]
        await destination_model_db.update_many(model_kinds)

        return destination

    async def snapshot(self, source: GraphName, label: str, timestamp: Optional[datetime] = None) -> GraphName:
        if not timestamp:
            timestamp = utc()

        if source.startswith("snapshot-"):
            raise ValueError("Can not snapshot a snapshot")

        time = utc_str(timestamp, date_format=UTC_Date_Format_short)
        check_graph_name(label)
        snapshot_name = GraphName(f"snapshot-{source}-{label}-{time}")
        return await self.copy(source, snapshot_name, replace_existing=False, validate_name=False, to_snapshot=True)

    async def delete(self, graph_name: GraphName) -> None:
        await self.db_access.delete_graph(graph_name)
        await self.db_access.delete_graph_model(graph_name)

    async def export_graph(self, graph_name: GraphName) -> AsyncIterator[str]:
        if not await self.db_access.db.has_graph(graph_name):
            raise ValueError(f"Graph {graph_name} does not exist")

        graph = cast(EventGraphDB, self.db_access.get_graph_db(graph_name)).real
        vertex_collection = graph_name
        default_edge_collection = graph.edge_collection(EdgeTypes.default)
        delete_edge_collection = graph.edge_collection(EdgeTypes.delete)
        model_collection = self.db_access.graph_model_name(graph_name)

        if not await self.db_access.db.has_collection(vertex_collection):
            model_collection = "model"

        async with self.db_access.db.begin_transaction(
            read=[vertex_collection, default_edge_collection, delete_edge_collection, model_collection]
        ) as tx:
            # Snapshot isolation ensures that the counts are consistent with the cursor data
            metadata = ExportMetadata(
                serializer_version="0.1.0",
                created_at=datetime.now().isoformat(),
                model_collection_size=await tx.count(model_collection),
                vertex_collection_size=await tx.count(vertex_collection),
                default_edge_collection_size=await tx.count(default_edge_collection),
                delete_edge_collection_size=await tx.count(delete_edge_collection),
            )

            # sturcture of the export file:
            # 1. metadata
            yield to_js_str(metadata)

            # 2. model collection
            cursor = AsyncCursor(await tx.all(model_collection), None)
            async for doc in cursor:
                yield dumps(doc)

            # 3. vertex collection
            cursor = AsyncCursor(await tx.all(vertex_collection), None)
            async for doc in cursor:
                yield dumps(doc)

            # 4. default edge collection
            cursor = AsyncCursor(await tx.all(default_edge_collection), None)
            async for doc in cursor:
                yield dumps(doc)

            # 5. delete edge collection
            cursor = AsyncCursor(await tx.all(delete_edge_collection), None)
            async for doc in cursor:
                yield dumps(doc)

    async def import_graph(self, graph_name: GraphName, stream: AsyncIterator[str], replace_existing: bool) -> None:
        if not self.lock:
            raise RuntimeError("GraphManager has not been started")

        async with self.lock:
            if await self.db_access.db.has_graph(graph_name):
                if replace_existing:
                    await self.delete(graph_name)
                else:
                    raise ValueError(f"Graph {graph_name} already exists")

            if await self.db_access.db.has_graph(graph_name):
                raise ValueError(f"Graph {graph_name} already exists")

            # temp graph to load the dump
            temp_graph = cast(
                EventGraphDB, await self.db_access.create_graph(GraphName(graph_name + "-temp"), validate_name=False)
            ).real

            metadata = from_js(
                loads(await stream.__anext__()), ExportMetadata  # pylint: disable=unnecessary-dunder-call
            )
            # check the serializer version, in the future we might need to support multiple versions
            if metadata.serializer_version != "0.1.0":
                raise ValueError(f"Unsupported dump version {metadata.serializer_version}")

            # import the model directly to the target graph model collection
            async def import_graph_model(data: AsyncIterator[str]) -> None:
                if metadata.model_collection_size == 0:
                    return

                position = 0
                kinds: List[Json] = []
                async for doc in data:
                    kinds.append(loads(doc))

                    # stop if we have reached the end of model
                    if position == metadata.model_collection_size - 1:
                        break
                    position += 1

                graph_model_db = await self.db_access.get_graph_model_db(graph_name)
                await graph_model_db.update_many(from_js(kinds, List[Kind]))

            # import the data into the temp graph
            async def import_buffered(data: AsyncIterator[str], doc_num: int, collection: str) -> None:
                if doc_num == 0:
                    return
                position = 0
                buffer = []
                async for doc in data:
                    # collect a batch
                    buffer.append(loads(doc))
                    # insert the batch
                    if len(buffer) == 1000:
                        await self.db_access.db.insert_many(collection, buffer)
                        buffer = []

                    if position == doc_num - 1:
                        break
                    position += 1

                if len(buffer) > 0:
                    await self.db_access.db.insert_many(collection, buffer)

            # step 1: import the graph model and the graph data into temporary collections
            await import_graph_model(stream)
            await import_buffered(stream, metadata.vertex_collection_size, temp_graph.vertex_name)
            await import_buffered(
                stream, metadata.default_edge_collection_size, temp_graph.edge_collection(EdgeTypes.default)
            )
            await import_buffered(
                stream, metadata.delete_edge_collection_size, temp_graph.edge_collection(EdgeTypes.delete)
            )

            # step 2: move the temporary graph collection to the final collection
            # we're using the copy to do an atomic rename and rewrite the edge references
            await self._copy_graph(temp_graph.vertex_name, graph_name, validate_name=False)

            # step 3: delete the temporary graph
            await self.db_access.delete_graph(temp_graph.name)


@frozen
class ExportMetadata:
    serializer_version: str
    created_at: str
    model_collection_size: int
    vertex_collection_size: int
    default_edge_collection_size: int
    delete_edge_collection_size: int


def _compress_timestamps(value: str) -> str:
    # support for negative years was removed in order to not parse the dashes in front
    iso8601 = re.compile(
        r"([\+]?\d{4}(?!\d{2}\b))((-?)((0[1-9]|1[0-2])(\3([12]\d|0[1-9]|3[01]))?|W([0-4]\d|5[0-2])(-?[1-7])?|(00[1-9]|0[1-9]\d|[12]\d{2}|3([0-5]\d|6[1-6])))([T\s]((([01]\d|2[0-3])((:?)[0-5]\d)?|24\:?00)([\.,]\d+(?!:))?)?(\17[0-5]\d([\.,]\d+)?)?([zZ]|([\+-])([01]\d|2[0-3]):?([0-5]\d)?)?)?)?"  # noqa: E501
    )

    # result string to be built
    result = ""
    # position in the source string
    source_pos = 0

    # find all timestamps in the source string
    for match in iso8601.finditer(value):
        # if there is a gap between the previous match and this one, copy the source string
        if match.start() > source_pos:
            result += value[source_pos : match.start()]

        # compact the timestamp
        timestamp = match.group(0)
        dt = parse(timestamp)
        compact_ts = utc_str(dt, UTC_Date_Format_short)
        result += compact_ts

        # update the pointer
        source_pos = match.end()

    # copy the rest of the source string
    if source_pos < len(value):
        result += value[source_pos:]

    return result



import re
from typing import Optional

from resotocore import version
from resotocore.model.model import (
    Kind,
    StringKind,
    BooleanKind,
    NumberKind,
    DateKind,
    DateTimeKind,
    DurationKind,
    TransformKind,
    DictionaryKind,
    AnyKind,
    ComplexKind,
    ArrayKind,
    Model,
    Property,
    any_kind,
)
from resotocore.types import Json
from resotolib.durations import DurationRe


def safe_name(kind: Kind) -> str:
    return re.sub("[^A-Za-z0-9]", "_", kind.fqn)


def simple_type(model: Model, kind: Kind) -> Optional[str]:
    if isinstance(kind, StringKind) and kind.fqn == "string":
        return "string"
    elif isinstance(kind, BooleanKind):
        return "boolean"
    elif isinstance(kind, NumberKind) and kind.fqn in ("int32", "int64"):
        return "integer"
    elif isinstance(kind, NumberKind) and kind.fqn in ("float", "double"):
        return "number"
    elif isinstance(kind, AnyKind):
        return "object"
    elif isinstance(kind, TransformKind):
        return simple_type(model, kind.source_kind or any_kind)
    else:
        return None


def ref_type(model: Model, kind: Kind) -> Json:
    if simple := simple_type(model, kind):
        return {"type": simple}
    elif isinstance(kind, ArrayKind):
        return {
            "type": "array",
            "items": ref_type(model, kind.inner),
        }
    elif isinstance(kind, DictionaryKind):
        return {
            "type": "object",
            "additionalProperties": ref_type(model, kind.value_kind),
        }
    else:
        return {"$ref": f"#/$defs/{safe_name(kind)}"}


def export_kind(model: Model, kind: Kind, is_base: bool) -> Optional[Json]:
    if isinstance(kind, (AnyKind, ArrayKind, DictionaryKind)):
        # no additional type information can be defined here
        return None
    elif isinstance(kind, StringKind):
        if kind.fqn != "string":
            result = {"type": "string", "additionalProperties": False}
            if kind.min_length:
                result["minLength"] = kind.min_length
            if kind.max_length:
                result["maxLength"] = kind.max_length
            if kind.pattern:
                result["pattern"] = kind.pattern
            if kind.enum:
                result["enum"] = kind.enum
            return result
        else:
            # build in type string does not need any additional information
            return None
    elif isinstance(kind, BooleanKind):
        if kind.fqn != "boolean":
            return {"type": "boolean", "additionalProperties": False}
        else:
            # build in type boolean does not need any additional information
            return None
    elif isinstance(kind, NumberKind):
        if kind.fqn not in ("int32", "int64", "float", "double"):
            result = {"additionalProperties": False}
            if kind.runtime_kind in ("int32", "int64"):
                result["type"] = "integer"
            elif kind.runtime_kind in ("float", "double"):
                result["type"] = "number"
            if kind.minimum:
                result["minimum"] = kind.minimum
            if kind.maximum:
                result["maximum"] = kind.maximum
            if kind.enum:
                result["enum"] = kind.enum
            return result
        else:
            # build in type number does not need any additional information
            return None
    elif isinstance(kind, DateKind):
        return {"type": "string", "format": "date"}
    elif isinstance(kind, DateTimeKind):
        return {"type": "string", "format": "date-time"}
    elif isinstance(kind, DurationKind):
        return {"type": "string", "pattern": DurationRe.pattern}
    elif isinstance(kind, TransformKind):
        assert kind.source_kind
        return export_kind(model, kind.source_kind, is_base)
    elif isinstance(kind, ComplexKind):
        ck: ComplexKind = kind

        def prop_model(p: Property) -> Json:
            pr = ref_type(model, ck.property_kind_of(p.name, any_kind))
            if desc := p.description:
                pr["description"] = desc
            return pr

        # we always need to define all properties.
        # Otherwise, it is not allowed to forbid additional properties
        all_props = {rp.name: prop_model(rp) for rp in ck.all_props()}
        # If this is a final kind, we expect a kind property with the name of the kind (constant)
        kind_prop = {} if is_base else {"kind": {"const": safe_name(ck)}}
        result = {
            "type": "object",
            "properties": all_props | kind_prop,
            "required": [prop.name for prop in kind.properties if prop.required],
            "additionalProperties": is_base | kind.allow_unknown_props,
        }
        if kind.bases:
            result["allOf"] = [ref_type(model, model[base]) for base in kind.bases]
        return result
    else:
        raise ValueError(f"Unknown kind: {kind}")


def json_schema(model: Model) -> Json:
    # filter out the predefined properties holder
    kinds = [k for k in model.kinds.values() if k.fqn != "predefined_properties"]
    all_bases = {b for k in kinds if isinstance(k, ComplexKind) for b in k.bases}
    return {
        "$id": f"https://resoto.com/schemas/{version()}/resources.json",
        "$schema": "https://json-schema.org/draft/2020-12/schema",  # latest draft
        "title": "Resoto Resource Model Schema",
        "oneOf": [
            ref_type(model, k)
            for k in kinds
            if isinstance(k, ComplexKind) and k.aggregate_root and k.fqn not in all_bases
        ],
        "$defs": {
            n: v
            for n, v in {safe_name(k): export_kind(model, k, k.fqn in all_bases) for k in kinds}.items()
            if v is not None
        },
    }



from attrs import define
from datetime import datetime
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import ArangoEntityDb
from resotocore.model.graph_access import DeferredEdge
from resotocore.ids import TaskId
from typing import List, cast
import logging

from resotocore.types import Json
from resotocore.ids import GraphName


@define
class PendingDeferredEdges:
    task_id: TaskId
    created_at: datetime  # update the corresponding TTL index when changing this name
    graph: GraphName
    edges: List[DeferredEdge]


TWO_HOURS = 7200


log = logging.getLogger(__name__)


class PendingDeferredEdgeDb(ArangoEntityDb[TaskId, PendingDeferredEdges]):
    async def create_update_schema(self) -> None:
        await super().create_update_schema()
        ttl_index_name = "deferred_edges_expiration_index"
        collection = self.db.collection(self.collection_name)
        if ttl_index_name not in {idx["name"] for idx in cast(List[Json], collection.indexes())}:
            log.info(f"Add index {ttl_index_name} on {collection.name}")
            collection.add_ttl_index(["created_at"], TWO_HOURS, "deferred_edges_expiration_index")


def pending_deferred_edge_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[TaskId, PendingDeferredEdges]:
    return PendingDeferredEdgeDb(db, collection, PendingDeferredEdges, lambda k: k.task_id)



from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, ArangoEntityDb
from resotocore.model.model import UsageDatapoint

ResourceUsageDb = EntityDb[str, UsageDatapoint]


class UsageDb(ArangoEntityDb[str, UsageDatapoint]):
    async def create_update_schema(self) -> None:
        name = self.collection_name
        db = self.db
        if not await db.has_collection(name):
            collecton = await db.create_collection(name)
            collecton.add_persistent_index(fields=["id", "at", "change_id"], storedValues=["v"])
            one_year = 60 * 60 * 24 * 365
            collecton.add_ttl_index(["at"], expiry_time=one_year)


def resource_usage_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[str, UsageDatapoint]:
    return UsageDb(db, collection, UsageDatapoint, lambda d: d.id + str(d.at))



from abc import ABC
from typing import Any, TypeVar

ServiceType = TypeVar("ServiceType", bound="Service")


class Service(ABC):
    async def start(self) -> Any:
        pass

    async def stop(self) -> None:
        pass

    async def __aenter__(self: ServiceType) -> ServiceType:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.stop()



import asyncio
import string
from abc import ABC, abstractmethod
from datetime import date, datetime, timedelta
from random import SystemRandom
from typing import List, Optional, Any, cast

from arango.database import StandardDatabase
from arango.typings import Json
from attrs import define
from networkx import MultiDiGraph
from pytest import mark, raises

from resotocore.analytics import CoreEvent, InMemoryEventSender
from resotocore.db.graphdb import ArangoGraphDB, GraphDB, EventGraphDB, HistoryChange
from resotocore.db.model import QueryModel, GraphUpdate
from resotocore.db.db_access import DbAccess
from resotocore.error import ConflictingChangeInProgress, NoSuchChangeError, InvalidBatchUpdate
from resotocore.ids import NodeId, GraphName
from resotocore.model.graph_access import GraphAccess, EdgeTypes, Section
from resotocore.model.model import Model, UsageDatapoint, UsageMetricValues
from resotocore.model.typed_model import from_js, to_js
from resotocore.query.model import Query, P, Navigation
from resotocore.query.query_parser import parse_query
from resotocore.types import JsonElement, EdgeType
from resotocore.util import AccessJson, utc, value_in_path, AccessNone


class BaseResource(ABC):
    def __init__(
        self,
        identifier: str,
    ) -> None:
        self.identifier = str(identifier)

    @abstractmethod
    def kind(self) -> str:
        pass


@define
class Foo(BaseResource):
    identifier: str
    name: Optional[str] = None
    some_int: int = 0
    some_string: str = "hello"
    now_is: datetime = utc()
    ctime: Optional[datetime] = None

    def kind(self) -> str:
        return "foo"


@define
class Inner:
    name: str
    inner: List["Inner"]


@define
class Bla(BaseResource):
    identifier: str
    name: Optional[str] = None
    now: date = date.today()
    f: int = 23
    g: Optional[List[int]] = None
    h: Optional[Inner] = None

    def __attrs_post_init__(self) -> None:
        self.g = self.g if self.g is not None else list(range(0, 5))
        if self.h is None:

            def nested(idx: int, level: int) -> Inner:
                return Inner(f"in_{level}_{idx}", [] if level == 0 else [nested(idx, level - 1) for idx in range(0, 2)])

            self.h = nested(0, 2)

    def kind(self) -> str:
        return "bla"


def create_graph(bla_text: str, width: int = 10) -> MultiDiGraph:
    graph = MultiDiGraph()

    def add_edge(from_node: str, to_node: str, edge_type: EdgeType = EdgeTypes.default) -> None:
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        graph.add_edge(from_node, to_node, key, edge_type=edge_type)

    def add_node(uid: str, kind: str, node: Optional[Json] = None, replace: bool = False) -> None:
        reported = {**(node if node else to_json(Foo(uid))), "kind": kind}
        graph.add_node(
            uid,
            id=uid,
            kinds=[kind],
            reported=reported,
            desired={"node_id": uid},
            metadata={"node_id": uid},
            replace=replace,
        )

    # root -> collector -> sub_root -> **rest
    add_node("root", "graph_root")
    add_node("collector", "cloud", replace=True)
    add_node("sub_root", "foo")
    add_edge("root", "collector")
    add_edge("collector", "sub_root")

    for o in range(0, width):
        oid = str(o)
        add_node(oid, "foo")
        add_edge("sub_root", oid)
        for i in range(0, width):
            iid = f"{o}_{i}"
            add_node(iid, "bla", node=to_json(Bla(iid, name=bla_text)))
            add_edge(oid, iid)
            add_edge(iid, oid, EdgeTypes.delete)
    return graph


def create_multi_collector_graph(width: int = 3) -> MultiDiGraph:
    graph = MultiDiGraph()

    def add_edge(from_node: str, to_node: str, edge_type: EdgeType = EdgeTypes.default) -> None:
        key = GraphAccess.edge_key(from_node, to_node, edge_type)
        graph.add_edge(from_node, to_node, key, edge_type=edge_type)

    def add_node(node_id: NodeId, kind: NodeId, replace: bool = False) -> NodeId:
        reported = {**to_json(Foo(node_id)), "id": node_id, "name": node_id, "kind": kind}
        graph.add_node(
            node_id,
            id=node_id,
            reported=reported,
            desired={},
            metadata={},
            hash="123",
            replace=replace,
            kind=kind,
            kinds=[kind],
            kinds_set={kind},
        )
        return node_id

    root = add_node(NodeId("root"), NodeId("graph_root"))
    for cloud_num in range(0, 2):
        cloud = add_node(NodeId(f"cloud_{cloud_num}"), NodeId("cloud"))
        add_edge(root, cloud)
        for account_num in range(0, 2):
            aid = f"{cloud_num}:{account_num}"
            account = add_node(NodeId(f"account_{aid}"), NodeId("account"))
            add_edge(cloud, account)
            add_edge(account, cloud, EdgeTypes.delete)
            for region_num in range(0, 2):
                rid = f"{aid}:{region_num}"
                region = add_node(NodeId(f"region_{rid}"), NodeId("region"), replace=True)
                add_edge(account, region)
                add_edge(region, account, EdgeTypes.delete)
                for parent_num in range(0, width):
                    pid = f"{rid}:{parent_num}"
                    parent = add_node(NodeId(f"parent_{pid}"), NodeId("parent"))
                    add_edge(region, parent)
                    add_edge(parent, region, EdgeTypes.delete)
                    for child_num in range(0, width):
                        cid = f"{pid}:{child_num}"
                        child = add_node(NodeId(f"child_{cid}"), NodeId("child"))
                        add_edge(parent, child)
                        add_edge(child, parent, EdgeTypes.delete)

    return graph


async def load_graph(db: GraphDB, model: Model, base_id: str = "sub_root") -> MultiDiGraph:
    blas = Query.by("foo", P("identifier") == base_id).traverse_out(0, Navigation.Max)
    return await db.search_graph(QueryModel(blas.on_section("reported"), model))


@mark.asyncio
async def test_update_merge_batched(graph_db: ArangoGraphDB, foo_model: Model, test_db: StandardDatabase) -> None:
    md = foo_model
    await graph_db.wipe()
    batch_id = "".join(SystemRandom().choice(string.ascii_letters) for _ in range(12))
    g = create_graph("yes or no")
    await graph_db.insert_usage_data(
        [
            UsageDatapoint("0", at=100, v={"cpu": UsageMetricValues(42, 42, 42)}, change_id=batch_id),
            UsageDatapoint("0", at=101, v={"cpu": UsageMetricValues(43, 43, 43)}, change_id="foo"),
        ]
    )

    # empty database: all changes are written to a temp table
    assert await graph_db.merge_graph(g, foo_model, batch_id, True) == (
        ["collector"],
        GraphUpdate(112, 1, 0, 212, 0, 0),
    )
    assert len((await load_graph(graph_db, md)).nodes) == 0
    # not allowed to commit an unknown batch
    with raises(NoSuchChangeError):
        await graph_db.commit_batch_update("does_not_exist")
    # commit the batch and see the changes reflected in the database
    await graph_db.commit_batch_update(batch_id)
    updated_graph = await load_graph(graph_db, md)
    assert len(updated_graph.nodes) == 111
    # ensure that all temp tables are removed
    assert len(list(filter(lambda c: c["name"].startswith("temp_"), cast(List[Json], test_db.collections())))) == 0
    # ensure the usage is there
    n = await graph_db.get_node(foo_model, NodeId("0")) or {}
    assert n.get("usage", {}).get("cpu") == {"min": 42, "avg": 42, "max": 42}
    # create a new batch that gets aborted: make sure all temp tables are gone
    batch_id = "will_be_aborted"
    await graph_db.merge_graph(g, foo_model, batch_id, True)
    await graph_db.abort_update(batch_id)
    assert len(list(filter(lambda c: c["name"].startswith("temp_"), cast(List[Json], test_db.collections())))) == 0


@mark.asyncio
async def test_merge_graph(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()

    def create(txt: str, width: int = 10) -> MultiDiGraph:
        return create_graph(txt, width=width)

    await graph_db.insert_usage_data(
        [
            UsageDatapoint("0", at=100, v={"cpu": UsageMetricValues(42, 42, 42)}, change_id="foo"),
            UsageDatapoint("0", at=101, v={"cpu": UsageMetricValues(43, 43, 43)}, change_id="bar"),
        ]
    )

    p = ["collector"]
    # empty database: all nodes and all edges have to be inserted, the root node is updated and the link to root added
    assert await graph_db.merge_graph(create("yes or no"), foo_model, maybe_change_id="foo") == (
        p,
        GraphUpdate(112, 1, 0, 212, 0, 0),
    )
    # check the usage
    n = await graph_db.get_node(foo_model, NodeId("0")) or {}
    assert n.get("usage", {}).get("cpu") == {"min": 42, "avg": 42, "max": 42}
    # exactly the same graph is updated: expect no changes
    assert await graph_db.merge_graph(create("yes or no"), foo_model) == (p, GraphUpdate(0, 0, 0, 0, 0, 0))
    # all bla entries have different content: expect 100 node updates, but no inserts or deletions
    assert await graph_db.merge_graph(create("maybe"), foo_model) == (p, GraphUpdate(0, 100, 0, 0, 0, 0))
    # the width of the graph is reduced: expect nodes and edges to be removed
    assert await graph_db.merge_graph(create("maybe", width=5), foo_model) == (p, GraphUpdate(0, 0, 80, 0, 0, 155))
    # going back to the previous graph: the same amount of nodes and edges is inserted
    assert await graph_db.merge_graph(create("maybe"), foo_model) == (p, GraphUpdate(80, 0, 0, 155, 0, 0))
    # updating with the same data again, does not perform any changes
    assert await graph_db.merge_graph(create("maybe"), foo_model) == (p, GraphUpdate(0, 0, 0, 0, 0, 0))


@mark.asyncio
async def test_merge_multi_graph(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()
    # nodes:
    # 2 collectors + 4 accounts + 8 regions + 24 parents + 72 children => 110 nodes to insert
    # 1 root which changes => 1 node to update
    # edges:
    # 110 default, 108 delete connections (missing: collector -> root) => 218 edge inserts
    nodes, info = await graph_db.merge_graph(create_multi_collector_graph(), foo_model)
    assert info == GraphUpdate(110, 1, 0, 218, 0, 0)
    assert len(nodes) == 8
    # doing the same thing again should do nothing
    nodes, info = await graph_db.merge_graph(create_multi_collector_graph(), foo_model)
    assert info == GraphUpdate(0, 0, 0, 0, 0, 0)
    assert len(nodes) == 8


@mark.asyncio
async def test_mark_update(filled_graph_db: ArangoGraphDB) -> None:
    db = filled_graph_db
    # make sure all changes are empty
    await db.db.truncate(db.in_progress)
    # change on 00 is allowed
    await db.mark_update(["00"], ["0", "sub_root", "root"], "update 00", False)
    # change on 01 is allowed
    await db.mark_update(["01"], ["0", "sub_root", "root"], "update 01", True)
    # same change id which tries to update the same subgraph root
    with raises(InvalidBatchUpdate):
        await db.mark_update(["01"], ["0", "sub_root", "root"], "update 01", True)
    # change on 0 is rejected, since there are changes "below" this node
    with raises(ConflictingChangeInProgress):
        await db.mark_update(["0"], ["sub_root"], "update 0 under node sub_root", False)
    # change on sub_root is rejected, since there are changes "below" this node
    with raises(ConflictingChangeInProgress):
        await db.mark_update(["sub_root"], ["root"], "update under node sub_root", False)
    # clean up for later tests
    await db.db.truncate(db.in_progress)


@mark.asyncio
async def test_query_list(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    blas = Query.by("foo", P("identifier") == "9").traverse_out().filter("bla", P("f") == 23)
    async with await filled_graph_db.search_list(QueryModel(blas.on_section("reported"), foo_model)) as gen:
        result = [from_js(x["reported"], Bla) async for x in gen]
        assert len(result) == 10

    foos_or_blas = parse_query("is([foo, bla])")
    async with await filled_graph_db.search_list(QueryModel(foos_or_blas.on_section("reported"), foo_model)) as gen:
        result = [x async for x in gen]
        assert len(result) == 111  # 113 minus 1 graph_root, minus one cloud


@mark.asyncio
async def test_query_not(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    # select everything that is not foo --> should be blas
    blas = Query.by(Query.mk_term("foo").not_term())
    async with await filled_graph_db.search_list(QueryModel(blas.on_section("reported"), foo_model)) as gen:
        result = [from_js(x["reported"], Bla) async for x in gen]
        assert len(result) == 102


@mark.asyncio
async def test_query_history(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    async def nodes(query: Query, **args: Any) -> List[Json]:
        async with await filled_graph_db.search_history(QueryModel(query, foo_model), **args) as crsr:
            return [x async for x in crsr]

    now = utc()
    five_min_ago = now - timedelta(minutes=5)
    assert len(await nodes(Query.by("foo"))) == 11
    assert len(await nodes(Query.by("foo"), after=five_min_ago)) == 11
    assert len(await nodes(Query.by("foo"), before=five_min_ago)) == 0
    assert len(await nodes(Query.by("foo"), after=five_min_ago, change=HistoryChange.node_created)) == 11
    assert len(await nodes(Query.by("foo"), after=five_min_ago, change=HistoryChange.node_deleted)) == 0


@mark.asyncio
async def test_query_graph(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    graph = await load_graph(filled_graph_db, foo_model)
    assert len(graph.edges) == 110
    assert len(graph.nodes.values()) == 111

    # filter data and tag result, and then traverse to the end of the graph in both directions
    around_me = Query.by("foo", P("identifier") == "9").tag("red").traverse_inout(start=0)
    graph = await filled_graph_db.search_graph(QueryModel(around_me.on_section("reported"), foo_model))
    assert len({x for x in graph.nodes}) == 12
    assert GraphAccess.root_id(graph) == "sub_root"
    assert list(graph.successors("sub_root"))[0] == "9"
    assert set(graph.successors("9")) == {f"9_{x}" for x in range(0, 10)}
    for from_node, to_node, data in graph.edges.data(True):
        assert from_node == "9" or to_node == "9"
        assert data == {"edge_type": "default"}

    node_id: str
    node: Json
    for node_id, node in graph.nodes.data(True):
        if node_id == "9":
            assert node["metadata"]["query_tag"] == "red"
        else:
            assert "tag" not in node["metadata"]

    async def assert_result(query: str, nodes: int, edges: int) -> None:
        q = parse_query(query)
        graph = await filled_graph_db.search_graph(QueryModel(q, foo_model))
        assert len(graph.nodes) == nodes
        assert len(graph.edges) == edges

    await assert_result("is(foo) and reported.identifier==9 <-delete[0:]default->", 11, 20)
    await assert_result("is(foo) and reported.identifier==9 <-default[0:]delete->", 4, 3)
    await assert_result("is(foo) and reported.identifier==9 <-default[0:]->", 14, 13)
    await assert_result("is(foo) and reported.identifier==9 <-delete[0:]->", 11, 10)
    await assert_result("is(foo) and reported.identifier==9 -default[0:]->", 11, 10)
    await assert_result("is(foo) and reported.identifier==9 <-delete[0:]-", 11, 10)
    await assert_result("is(foo) and reported.identifier==9 <-default[0:]-", 4, 3)
    await assert_result("is(foo) and reported.identifier==9 -delete[0:]->", 1, 0)


@mark.asyncio
async def test_query_nested(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    async def assert_count(query: str, count: int) -> None:
        q = parse_query(query).on_section("reported")
        async with await filled_graph_db.search_list(QueryModel(q, foo_model), with_count=True) as gen:
            assert gen.cursor.count() == count
            assert len([a async for a in gen]) == count

    await assert_count("is(bla) and h.inner[*].inner[*].name=in_0_1", 100)
    await assert_count("is(bla) and h.inner[*].inner[*].inner == []", 100)
    await assert_count("is(bla) and g[*] = 2", 100)
    await assert_count("is(bla) and g any = 2", 100)
    await assert_count("is(bla) and g all = 2", 0)
    await assert_count("is(bla) and g none = 2", 0)
    await assert_count("is(bla) and g[*] any = 2", 100)
    await assert_count("is(bla) and g[*] all = 2", 0)
    await assert_count("is(bla) and g[*] none = 2", 0)


@mark.asyncio
async def test_query_aggregate(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    agg_query = parse_query("aggregate(kind: count(identifier) as instances): is(foo)").on_section("reported")
    async with await filled_graph_db.search_aggregation(QueryModel(agg_query, foo_model)) as gen:
        assert [x async for x in gen] == [{"group": {"kind": "foo"}, "instances": 11}]

    agg_combined_var_query = parse_query(
        'aggregate("test_{kind}_{some_int}_{does_not_exist}" as kind: count(identifier) as instances): is("foo")'
    ).on_section("reported")

    async with await filled_graph_db.search_aggregation(QueryModel(agg_combined_var_query, foo_model)) as g:
        assert [x async for x in g] == [{"group": {"kind": "test_foo_0_"}, "instances": 11}]

    agg_multi_fn_same_prop = parse_query('aggregate(sum(f) as a, max(f) as b): is("bla")').on_section("reported")
    async with await filled_graph_db.search_aggregation(QueryModel(agg_multi_fn_same_prop, foo_model)) as g:
        assert [x async for x in g] == [{"a": 2300, "b": 23}]


@mark.asyncio
async def test_query_with_fulltext(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    async def search(query: str) -> List[JsonElement]:
        async with await filled_graph_db.search_list(QueryModel(parse_query(query), foo_model)) as cursor:
            return [elem async for elem in cursor]

    # Note: the fulltext index is eventually consistent. Since the database is wiped and cleaned for this test
    # we should not have any assumptions about the results, other than the query succeeds
    await search('(("a" and "b") or ("c" and "d")) and "e"')
    await search('is(foo) and "test" --> "bim bam bom bum"')
    await search('is(foo) {a: --> "some prop" } "some other prop" --> "bim bam bom bum"')


@mark.asyncio
async def test_query_merge(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    q = parse_query(
        "is(foo) --> is(bla) { "
        "foo.bar.parents[]: <-[1:]-, "
        "foo.child: -->, "
        "walk: <-- -->, "
        "bla.agg: aggregate(sum(1) as count): <-[0:]- "
        "}"
    )
    async with await filled_graph_db.search_list(QueryModel(q, foo_model), with_count=True) as cursor:
        assert cursor.count() == 100
        async for bla in cursor:
            b = AccessJson(bla)
            assert b.reported.kind == "bla"
            assert len(b.foo.bar.parents) == 4
            for parent in b.foo.bar.parents:
                assert parent.reported.kind in ["foo", "cloud", "graph_root"]
            assert b.walk.reported.kind == "bla"
            assert b.foo.child == AccessNone()
            assert b.bla.agg == [{"count": 5}]


@mark.asyncio
async def test_query_with_clause(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    async def query(q: str) -> List[Json]:
        agg_query = parse_query(q)
        async with await filled_graph_db.search_list(QueryModel(agg_query.on_section("reported"), foo_model)) as cursor:
            return [bla async for bla in cursor]

    assert len(await query("is(bla) with(any, <-- is(foo))")) == 100
    assert len(await query('is(bla) with(any, <-- is(foo) and identifier=~"1")')) == 10
    assert len(await query("is(bla) with(empty, <-- is(foo))")) == 0
    assert len(await query("is(bla) with(any, <-- is(bla))")) == 0
    assert len(await query("is(bla) with(empty, <-- is(bla))")) == 100
    assert len(await query('is(bla) with(count==1, <-- is(foo) and identifier=~"1")')) == 10
    assert len(await query('is(bla) with(count==2, <-- is(foo) and identifier=~"1")')) == 0
    assert len(await query("is(bla) with(any, <-- with(any, <-- is(foo)))")) == 100


@mark.asyncio
async def test_no_null_if_undefined(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()
    # imported graph should not have any desired or metadata sections
    graph = create_graph("test", 0)
    for _, node in graph.nodes(True):
        del node["desired"]
        del node["metadata"]
    await graph_db.merge_graph(graph, foo_model)
    async with await graph_db.search_list(QueryModel(parse_query("all"), foo_model)) as cursor:
        async for elem in cursor:
            assert "reported" in elem
            assert "desired" not in elem
            assert "metadata" not in elem


@mark.asyncio
async def test_get_node(filled_graph_db: ArangoGraphDB, foo_model: Model) -> None:
    # load sub_root as foo
    sub_root = to_foo(await filled_graph_db.get_node(foo_model, NodeId("sub_root")))
    assert sub_root is not None
    assert isinstance(sub_root, Foo)
    # load node 7 as foo
    node_7_json = await filled_graph_db.get_node(foo_model, NodeId("7"))
    node_7 = to_foo(node_7_json)
    assert node_7 is not None
    assert isinstance(node_7, Foo)
    # make sure that all synthetic properties are rendered (the age should not be older than 1 second => 0s or 1s)
    assert node_7_json[Section.reported]["age"] in ["0s", "1s"]  # type: ignore
    # load node 1_2 as bla
    node_1_2 = to_bla(await filled_graph_db.get_node(foo_model, NodeId("1_2")))
    assert node_1_2 is not None
    assert isinstance(node_1_2, Bla)


@mark.asyncio
async def test_insert_node(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()
    json = await graph_db.create_node(
        foo_model, NodeId("some_new_id"), to_json(Foo("some_new_id", "name")), NodeId("root")
    )
    assert to_foo(json).identifier == "some_new_id"
    assert to_foo(await graph_db.get_node(foo_model, NodeId("some_new_id"))).identifier == "some_new_id"


@mark.asyncio
async def test_update_node(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()
    await graph_db.create_node(foo_model, NodeId("some_other"), to_json(Foo("some_other", "foo")), NodeId("root"))
    json_patch = await graph_db.update_node(foo_model, NodeId("some_other"), {"name": "bla"}, False, "reported")
    assert to_foo(json_patch).name == "bla"
    assert to_foo(await graph_db.get_node(foo_model, NodeId("some_other"))).name == "bla"
    json_replace = (
        await graph_db.update_node(
            foo_model, NodeId("some_other"), {"kind": "bla", "identifier": "123"}, True, "reported"
        )
    )["reported"]
    json_replace.pop("ctime")  # ctime is added by the system automatically. remove it
    assert json_replace == {"kind": "bla", "identifier": "123"}


@mark.asyncio
async def test_update_nodes(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    def expect(jsons: List[Json], path: List[str], value: JsonElement) -> None:
        for js in jsons:
            v = value_in_path(js, path)
            assert v is not None
            assert v == value

    await graph_db.wipe()
    await graph_db.create_node(foo_model, NodeId("id1"), to_json(Foo("id1", "foo")), NodeId("root"))
    await graph_db.create_node(foo_model, NodeId("id2"), to_json(Foo("id2", "foo")), NodeId("root"))
    # only change the desired section
    change1 = {"desired": {"test": True}}
    result1 = [a async for a in graph_db.update_nodes(foo_model, {NodeId("id1"): change1, NodeId("id2"): change1})]
    assert len(result1) == 2
    expect(result1, ["desired", "test"], True)
    # only change the metadata section
    change2 = {"metadata": {"test": True}}
    result2 = [a async for a in graph_db.update_nodes(foo_model, {NodeId("id1"): change2, NodeId("id2"): change2})]
    assert len(result2) == 2
    expect(result2, ["metadata", "test"], True)
    # change all sections including the reported section
    change3 = {"desired": {"test": False}, "metadata": {"test": False}, "reported": {"name": "test"}}
    node_raw_id1 = AccessJson.wrap_object(graph_db.db.db.collection(graph_db.name).get("id1"))
    result3 = [a async for a in graph_db.update_nodes(foo_model, {NodeId("id1"): change3, NodeId("id2"): change3})]
    assert len(result3) == 2
    expect(result3, ["desired", "test"], False)
    expect(result3, ["metadata", "test"], False)
    expect(result3, ["reported", "name"], "test")
    # make sure the db is updated
    node_raw_id1_updated = AccessJson.wrap_object(graph_db.db.db.collection(graph_db.name).get("id1"))
    assert node_raw_id1.reported.name != node_raw_id1_updated.reported.name
    assert node_raw_id1.desired.test != node_raw_id1_updated.desired.test
    assert node_raw_id1.metadata.test != node_raw_id1_updated.metadata.test
    assert node_raw_id1.flat != node_raw_id1_updated.flat
    assert node_raw_id1.hash != node_raw_id1_updated.hash
    assert "test" in node_raw_id1_updated.flat
    change4 = {"desired": None, "metadata": None}
    result4 = [
        a
        async for a in graph_db.update_nodes(foo_model, {NodeId("id1"): change4.copy(), NodeId("id2"): change4.copy()})
    ]
    assert len(result4) == 4
    assert all("desired" not in a for a in result4)
    assert len([a for a in result4 if "metadata" not in a]) == 2


@mark.asyncio
async def test_delete_node(graph_db: ArangoGraphDB, foo_model: Model) -> None:
    await graph_db.wipe()
    await graph_db.create_node(foo_model, NodeId("sub_root"), to_json(Foo("sub_root", "foo")), NodeId("root"))
    await graph_db.create_node(
        foo_model, NodeId("some_other_child"), to_json(Foo("some_other_child", "foo")), NodeId("sub_root")
    )
    await graph_db.create_node(foo_model, NodeId("born_to_die"), to_json(Foo("born_to_die", "foo")), NodeId("sub_root"))
    await graph_db.delete_node(NodeId("born_to_die"))
    assert await graph_db.get_node(foo_model, NodeId("born_to_die")) is None
    with raises(AttributeError) as not_allowed:
        await graph_db.delete_node(NodeId("sub_root"))
    assert str(not_allowed.value) == "Can not delete node, since it has 1 child(ren)!"


@mark.asyncio
async def test_events(
    event_graph_db: EventGraphDB, foo_model: Model, event_sender: InMemoryEventSender, db_access: DbAccess
) -> None:
    await event_graph_db.create_node(foo_model, NodeId("some_other"), to_json(Foo("some_other", "foo")), NodeId("root"))
    await event_graph_db.update_node(foo_model, NodeId("some_other"), {"name": "bla"}, False, "reported")
    await event_graph_db.delete_node(NodeId("some_other"))
    await event_graph_db.merge_graph(create_graph("yes or no", width=1), foo_model)
    await event_graph_db.merge_graph(create_graph("maybe", width=1), foo_model, "batch1", True)
    copy_graph_name = GraphName("graph_copy_for_event")
    await db_access.delete_graph(copy_graph_name)
    await event_graph_db.copy_graph(copy_graph_name)
    # make sure all events will arrive
    await asyncio.sleep(0.1)
    # ensure the correct count and order of events
    assert [a.kind for a in event_sender.events] == [
        CoreEvent.NodeCreated,
        CoreEvent.NodeUpdated,
        CoreEvent.NodeDeleted,
        CoreEvent.GraphMerged,
        CoreEvent.BatchUpdateGraphMerged,
        CoreEvent.GraphCopied,
    ]
    merge_event = AccessJson(event_sender.events[3].context)
    assert merge_event.graph == event_graph_db.graph_name
    assert merge_event.providers == ["collector"]
    assert merge_event.batch is False
    copy_event = AccessJson(event_sender.events[5].context)
    assert copy_event.graph == event_graph_db.name
    assert copy_event.to_graph == "graph_copy_for_event"
    await db_access.delete_graph(copy_graph_name)


@mark.asyncio
async def test_db_copy(graph_db: ArangoGraphDB, foo_model: Model, db_access: DbAccess) -> None:
    await graph_db.wipe()

    # populate some data in the graphes
    nodes, info = await graph_db.merge_graph(create_multi_collector_graph(), foo_model)
    assert info == GraphUpdate(110, 1, 0, 218, 0, 0)
    assert len(nodes) == 8

    db = graph_db.db
    copy_db_name = GraphName("copy_" + graph_db.name)
    # make sure the copy graph does not exist
    await db_access.delete_graph(copy_db_name)

    # copy the graph
    copy_db = await graph_db.copy_graph(copy_db_name)
    assert copy_db.name == copy_db_name

    async def validate(original_db_name: GraphName, copy_db_name: str) -> None:
        # validate the vertices
        existing_vertex_ids = {a["_key"] for a in await db.all(original_db_name)}
        copy_vertex_ids = {a["_key"] for a in await db.all(copy_db_name)}
        assert existing_vertex_ids == copy_vertex_ids

        # validate the default edges
        existing_default_edge_ids = {a["_key"] for a in await db.all(f"{original_db_name}_default")}
        copy_default_edge_ids = {a["_key"] for a in await db.all(f"{copy_db_name}_default")}
        assert existing_default_edge_ids == copy_default_edge_ids

        # validate the delete edges
        existing_delete_edge_ids = {a["_key"] for a in await db.all(f"{original_db_name}_delete")}
        copy_delete_edge_ids = {a["_key"] for a in await db.all(f"{copy_db_name}_delete")}
        assert existing_delete_edge_ids == copy_delete_edge_ids

    await validate(graph_db.name, copy_db.name)

    # check snapshots
    snapshot_db_name = GraphName("snapshot-" + graph_db.name)
    snapshot_db = await graph_db.copy_graph(snapshot_db_name, to_snapshot=True)
    assert snapshot_db.name == snapshot_db_name
    await validate(graph_db.name, snapshot_db.name)

    # clean up
    await snapshot_db.wipe()


@mark.asyncio
async def test_no_snapshot_usage(graph_db: ArangoGraphDB, foo_model: Model, db_access: DbAccess) -> None:
    await graph_db.wipe()
    await db_access.delete_graph(GraphName("snapshot-" + graph_db.name))

    snapshot_db_name = GraphName("snapshot-" + graph_db.name)
    snapshot_db = await graph_db.copy_graph(snapshot_db_name, to_snapshot=True)

    with raises(ValueError) as ex:
        await snapshot_db.insert_usage_data(
            [UsageDatapoint("foo", 42, "foo", {"cpu": UsageMetricValues(0.42, 0.42, 0.42)})]
        )

    assert str(ex.value) == "Cannot insert usage data into a snapshot graph"

    # clean up
    await snapshot_db.wipe()


def test_render_metadata_section(foo_model: Model) -> None:
    printer = ArangoGraphDB.document_to_instance_fn(foo_model)
    out = printer({"_key": "1", "reported": {"kind": "foo"}, "metadata": {"exported_at": "2023-03-06T19:37:51Z"}})
    assert "exported_age" in out["metadata"]  # exported_age is not part of the document, but should be added


def to_json(obj: BaseResource) -> Json:
    return {"kind": obj.kind(), **to_js(obj)}


def to_bla(json: Optional[Json]) -> Bla:
    assert json is not None
    return from_js(json["reported"], Bla)


def to_foo(json: Optional[Json]) -> Foo:
    assert json is not None
    return from_js(json["reported"], Foo)



import string
from datetime import datetime
from typing import TypeVar, Callable, Any, cast, Optional, List, Generator

from aiostream import stream
from aiostream.core import Stream
from hypothesis.strategies import (
    SearchStrategy,
    just,
    text,
    dictionaries,
    booleans,
    integers,
    lists,
    composite,
    sampled_from,
)

from resotocore.model.resolve_in_graph import NodePath
from resotocore.types import JsonElement, Json
from resotocore.util import value_in_path, interleave

T = TypeVar("T")
UD = Callable[[SearchStrategy[Any]], Any]


def optional(st: SearchStrategy[T]) -> SearchStrategy[Optional[T]]:
    return st | just(None)


class Drawer:
    """
    Only here for getting a drawer for typed drawings.
    """

    def __init__(self, hypo_drawer: Callable[[SearchStrategy[Any]], Any]):
        self._drawer = hypo_drawer

    def draw(self, st: SearchStrategy[T]) -> T:
        return cast(T, self._drawer(st))

    def optional(self, st: SearchStrategy[T]) -> Optional[T]:
        return self.draw(optional(st))


any_ws_digits_string = text(alphabet=string.ascii_letters + string.whitespace + string.digits, min_size=0, max_size=10)
any_string = text(alphabet=string.ascii_letters, min_size=3, max_size=10)
kind_gen = sampled_from(["volume", "instance", "load_balancer", "volume_type"])

any_datetime = integers(min_value=0, max_value=1688108028).map(lambda x: datetime.fromtimestamp(x))


@composite
def json_element_gen(ud: UD) -> JsonElement:
    return cast(JsonElement, ud(json_object_gen | json_simple_element_gen | json_array_gen))


json_simple_element_gen = any_ws_digits_string | booleans() | integers(min_value=0, max_value=100000) | just(None)
json_object_gen = dictionaries(any_string, json_element_gen(), min_size=1, max_size=5)
json_array_gen = lists(json_element_gen(), max_size=5)


@composite
def node_gen(ud: UD) -> Json:
    d = Drawer(ud)
    uid = d.draw(any_string)
    name = d.draw(any_string)
    kind = d.draw(kind_gen)
    reported = d.draw(json_object_gen)
    metadata = d.draw(json_object_gen)
    desired = d.draw(json_object_gen)
    return {
        "id": uid,
        "kinds": [kind],
        "reported": {**reported, "kind": kind, "id": uid, "name": name},
        "metadata": metadata,
        "desired": desired,
        "type": "node",
    }


def graph_stream(node_list: List[Json]) -> Stream:
    def from_node() -> Generator[Json, Any, None]:
        for node in node_list:
            yield node
        node_ids = [value_in_path(a, NodePath.node_id) for a in node_list]
        for from_n, to_n in interleave(node_ids):
            yield {"type": "edge", "from": from_n, "to": to_n}

    return stream.iterate(from_node())
